/*  cpse-cli.c -- command-line interface to CPSE
    Copyright (C) 2008  SEIKO EPSON CORPORATION

    This file is part of "Custom Paper Size Editor".

    "Custom Paper Size Editor" is free software.
    You can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License or at your option any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY;  without even the implied warranty of FITNESS
    FOR A PARTICULAR PURPOSE or MERCHANTABILITY.
    See the GNU General Public License for more details.

    You should have received a verbatim copy of the GNU General Public
    License along with this program; if not, write to:

        Free Software Foundation, Inc.
        59 Temple Place, Suite 330
        Boston, MA  02111-1307  USA
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ctype.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>

#include "generix.h"
#include "message.h"
#include "model.h"
#include "ppd.h"

/* Program-wide reusable buffer for line input via getline().
 */
static gchar *line_buffer = NULL;
static size_t line_length = 0;


static const gchar *prompt = " > ";

/* Traditional console dimensions */
#define SCREEN_COLS	80
#define SCREEN_ROWS	24

static void
display_screen (const gchar *action, const gchar *title, const gchar *header,
		const gchar **lines, size_t count)
{
  gint blanks = SCREEN_ROWS - 3 - count;

  if (0 > blanks) blanks = 0;

  printf ("\n");
  if (!title)
    printf (PACKAGE_NAME);
  else
    printf (title);
  printf ("\n");
  if (header) printf (header);
  printf ("\n");

  while (count--)
    {
      printf (*lines++);
      printf ("\n");
    }
  while (blanks--)
    {
      printf ("\n");
    }
  printf (action);
  printf (prompt);
}

/*! Removes trailing new lines from a string \a s.

    Returns the modified string.  Both the '\n' and '\r' characters
    are considered a newline character.
 */
static gchar *
chomp (gchar *s)
{
  gchar *_s;

  if (s)
    while ((_s = strrchr (s, '\n')) || (_s = strrchr (s, '\r')))
      *_s = '\0';

  return (s);
}

/*! Returns a pointer to the first non-whitespace character in \a s.
 */
static gchar *
skip_space (const gchar *s)
{
  const gchar *p = s;
  gchar *locale = setlocale (LC_CTYPE, "C");

  while (p && *p && isspace (*p))
    ++p;

  setlocale (LC_CTYPE, locale);
  return ((gchar *) p);
}

typedef enum {
  ERROR = -1,
  SELECT_PPDFILE,
  SELECT_PRINTER,
  SELECT_UNITS,
  ACTION_EDIT,
  ACTION_SAVE,
  ACTION_APPLY,
  CUSTOM_ADD,
  CUSTOM_CHANGE,
  CUSTOM_REMOVE,
  QUIT,
} action;

struct cmd
{
	gchar  command;
       action  action_id;
  const gchar *description;
};

/*! Computes the number of elements in a "NULL" terminated array.
 */
#define array_size(array,type)	(sizeof (array) / sizeof (type) - 1)


/* Command menus for all levels.
   Entries are in order of output and command letters need to be
   unique within a single menu and uppercase.
   All menus are "NULL" terminated.

   NOTE: if all actions can be caught in a single function signature I
   will probably replace the action_id with an action_fp.  That way we
   no longer need the ugly switch in main().
 */
static struct cmd menu_start[] = {
  {'P', SELECT_PRINTER, N_("Select Printer")},
  {'F', SELECT_PPDFILE, N_("Select PPD File")},
  {'Q', QUIT, N_("Quit")},
  { 0 , ERROR, NULL }
};

static struct cmd menu_act[] = {
  {'E', ACTION_EDIT, N_("Edit")},
  {'S', ACTION_SAVE, N_("Save")},
  {'A', ACTION_APPLY, N_("Apply to Printer")},
  {'U', SELECT_UNITS, N_("Units")},
  {'Q', QUIT, N_("Quit")},
  { 0 , ERROR, NULL }
};

static struct cmd menu_edit[] = {
  {'A', CUSTOM_ADD, N_("Add")},
  {'C', CUSTOM_CHANGE, N_("Change")},
  {'R', CUSTOM_REMOVE, N_("Remove")},
  {'U', SELECT_UNITS, N_("Units")},
  {'F', QUIT, N_("Finish")},
  { 0 , ERROR, NULL }
};

/* All menus in order of their (implicitly) associated level.
 */
static struct cmd *menu[] = {
  menu_start,
  menu_act,
  menu_edit,
  NULL
};

/* Not really an action menu, but let's reuse the infrastructure for
   unit selection.
 */
static struct cmd unit[] = {
  {'C', UNIT_CM  , N_("cm")},
  {'I', UNIT_INCH, N_("inch")},
#if ENABLE_EXTRA_UNITS
  {'M', UNIT_MM  , N_("mm")},
  {'D', UNIT_DOTS, N_("dots")},
#endif
  { 0 , ERROR, NULL }
};

static const gchar *
find_units_string (gint units)
{
  struct cmd *submenu = unit;

  while (submenu && submenu->command && units != submenu->action_id)
    ++submenu;

  if (submenu->command)
    return (submenu->description);

  return ("dpi");
}

static struct cmd *
get_menu (size_t level)
{
  require (0 < level && level <= array_size (menu, struct cmd *));
  return (menu[level-1]);
}

static gchar *
get_menu_line (const struct cmd *submenu)
{
  gchar *buffer = NULL;
  size_t length = 0;

  if (!submenu)
    return (NULL);

  buffer = (gchar *) g_malloc (sizeof (gchar));
  if (!buffer)
    return (NULL);

  buffer[0] = '\0';
  while (submenu && submenu->command)
    {
      gint _length = (strlen (buffer) + strlen (" (")
		      + strlen (&submenu->command) + strlen (") ")
		      + strlen (_(submenu->description))
		      + strlen (" /") + 1);
      gchar *_buffer;

      _buffer = (gchar *) g_realloc (buffer, _length);
      if (!_buffer)
	{
	  g_free (buffer);
	  return (NULL);
	}
      buffer = _buffer;
      length = _length;

      strcat (buffer, " (");
      strcat (buffer, &(submenu->command));
      strcat (buffer, ") ");
      strcat (buffer, _(submenu->description));
      strcat (buffer, " /");
      ++submenu;
    }
  {				/* "zap" the trailing " /" bit */
    gchar *p = strrchr (buffer, ' ');
    if (p) *p = '\0';
  }
  return (buffer);
}

static const gchar *
get_menu_prompt (size_t level)
{
  static gchar **prompt = NULL;
  static size_t  count  = array_size (menu, struct cmd *);

  require (0 < level && level <= count);

  if (!prompt)			/* lazy initialisation */
    {
      gint i;

      prompt = (gchar **) g_malloc0 (count * sizeof (gchar *));
      if (!prompt)
	{
	  log_errno (ENOMEM);
	  exit (EXIT_FAILURE);
	}

      for (i = 0; i < count; ++i)
	{
	  prompt[i] = get_menu_line (menu[i]);
	  if (!prompt[i])
	    {
	      while (i) g_free (prompt[--i]);
	      g_free (prompt);
	      log_errno (ENOMEM);
	      exit (EXIT_FAILURE);
	    }
	}
    }
  return (prompt[level-1]);
}

static action
parse (const struct cmd *submenu, const gchar *buffer, size_t length)
{
  action a = ERROR;
  gchar *locale = NULL;

  require (submenu);
  locale  = setlocale (LC_CTYPE, "C");

  while (buffer && 0 < length && isspace (*buffer))
    ++buffer, --length;

  if (0 < length)
    {
      gchar command = toupper (*buffer);

      while (submenu && submenu->command && command != submenu->command)
	++submenu;
      if (submenu->command)
	a = submenu->action_id;
    }

  setlocale (LC_CTYPE, locale);
  return (a);
}

static void
display_printer_menu (const CPSEModel *model)
{
  const gchar *fmt = "%3d %-75.75s";
  gchar action[SCREEN_COLS];
  gchar header[SCREEN_COLS];
  gchar **printer = NULL;
  size_t i = 0;

  require (model);

  if (0 >= model->dest_sz)
    {
      log_msg ("No printers found");
      return;
    }

  /* Allocate all the memory we will be needing ahead of time.
   */
  printer = (gchar **) g_malloc (model->dest_sz * sizeof (gchar *));
  if (!printer)
    {
      log_errno (ENOMEM);
      return;
    }
  for (i = 0; i < model->dest_sz; ++i)
    {
      printer[i] = (gchar *) g_malloc (SCREEN_COLS * sizeof (gchar));
      if (!printer[i])
	{
	  while (i--)
	    g_free (printer[i]);
	  g_free (printer);
	  log_errno (ENOMEM);
	  return;
	}
      printer[i][0] = '\0';
    }

  for (i = 0; i < model->dest_sz; ++i)
    {
      snprintf (printer[i], SCREEN_COLS, fmt, i+1, model->dest[i].name);
    }
  snprintf (header, SCREEN_COLS, "%s %s", _("No."), _("Printer"));
  snprintf (action, SCREEN_COLS, _("Select Printer (1-%d) / Cancel (C)"),
	    model->dest_sz);
  display_screen (action, NULL, header, (const gchar **) printer,
		  model->dest_sz);

  while (i--)
    g_free (printer[i]);
  g_free (printer);
}

static void
display_paper_menu (size_t level, const CPSEModel *model)
{
  const size_t UNIT_COLS = 20;	/* sync with third arg in hdr */

  const gchar *ttl = NULL;
  const gchar *hdr = "%-3.3s %-32.32s %19.19s %8.8s %8.8s %-4.4s";
  const gchar *fmt = "%3d %-52.52s %8.2f %8.2f";
  const gchar *filename = NULL;
  gchar title[SCREEN_COLS];
  gchar header[SCREEN_COLS];
  gchar units[UNIT_COLS];
  gchar **paper = NULL;
  size_t i = 0;
  size_t count;

  require (model);

  /* Allocate all the memory we will be needing ahead of time.
   */
  count = cpse_model_paper_count (model);
  paper = (gchar **) g_malloc (count * sizeof (gchar *));
  if (!paper)
    {
      log_errno (ENOMEM);
      return;
    }
  for (i = 0; i < count; ++i)
    {
      paper[i] = (gchar *) g_malloc (SCREEN_COLS * sizeof (gchar));
      if (!paper[i])
	{
	  while (i--)
	    g_free (paper[i]);
	  g_free (paper);
	  log_errno (ENOMEM);
	  return;
	}
      paper[i][0] = '\0';
    }

  if (model->printer.ppd_file)
    {				/* get basename */
      filename = strrchr (model->printer.ppd_file, G_DIR_SEPARATOR);
      if (filename)
	++filename;
      else
	filename = model->printer.ppd_file;
    }
  if (model->printer.name && filename)
    {
      ttl = "%s: %s / %s: %s";
      snprintf (title, SCREEN_COLS, ttl, _("Printer"), model->printer.name,
		_("File"), filename);
    }
  else
    {
      ttl = "%s: %s";
      if (filename)
	snprintf (title, SCREEN_COLS, ttl, _("File"), filename);
      else
	snprintf (title, SCREEN_COLS, ttl, _("Printer"), model->printer.name);
    }

  for (i = 0; i < count; ++i)
    {
      snprintf (paper[i], SCREEN_COLS, fmt, i+1,
		model->custom_paper[i].text,
		cpse_model_convert (model->custom_paper[i].width , UNIT_DOTS,
				    model->units),
		cpse_model_convert (model->custom_paper[i].height, UNIT_DOTS,
				    model->units));
      if (model->has_hw_res)
	{			/* abuse header as a temporary buffer */
	  memcpy (header, paper[i], SCREEN_COLS);
	  snprintf (paper[i], SCREEN_COLS, "%s %4.0f", header,
		    model->custom_paper[i].resolution);
	}
    }

  snprintf (units, UNIT_COLS, "[%s: %s]", _("Units"),
	    _(find_units_string (model->units)));
  snprintf (header, SCREEN_COLS, hdr, _("No."), _("Name"), units, _("Width"),
	    _("Height"), (model->has_hw_res ? _("Res.") : ""));

  display_screen (get_menu_prompt (level), title, header,
		  (const gchar **) paper, count);

  while (i--)
    g_free (paper[i]);
  g_free (paper);
}

static action
get_action (size_t level, const CPSEModel *model)
{
  require (model);

  if (1 < level)
    display_paper_menu (level, model);
  else
    display_screen (get_menu_prompt (level), NULL, NULL, NULL, 0);

  if (-1 == getline (&line_buffer, &line_length, stdin))
    {
      return (ERROR);
    }

  return (parse (get_menu (level), line_buffer, line_length));
}

static gboolean
in_range (gdouble value, const CPSEValue *range, gdouble fuzz)
{
  require (range);

  return (range->min - fuzz <= value && value <= range->max + fuzz);
}

static gboolean
read_value (gdouble *v, const CPSEValue *range, const gchar *description,
	    gint units, gboolean allow_cancel, const gchar *format)
{
  const gdouble epsilon = (72.0 / 100 + 0.01) / 2;
  gboolean cancelled = FALSE;
  gdouble  value;

  require (v);
  require (range);
  require (description);
  require (format);

  do
    {
      const gchar *p;

      if (allow_cancel)
	{
	  printf (" ");
	  printf (format, range->min, range->max, description);
	  printf (" / (C) %s%s", _("Cancel"), prompt);
	}
      else
	{
	  printf (" ");
	  printf (format, description,
		  cpse_model_convert (range->min, UNIT_DOTS, units),
		  cpse_model_convert (range->max, UNIT_DOTS, units),
		  cpse_model_convert (range->val, UNIT_DOTS, units),
		  _(find_units_string (units)));
	  printf (prompt);
	}

      if (-1 == getline (&line_buffer, &line_length, stdin))
	return (cancelled = TRUE);

      p = skip_space (chomp (line_buffer));
      if (0 == strlen (p))
	value = range->val;
      else
	{
	  if (allow_cancel)
	    {
	      gchar *locale = setlocale (LC_CTYPE, "C");
	      cancelled = (isalpha (*p) && 'C' == toupper (*p));
	      setlocale (LC_CTYPE, locale);
	    }
	  if (!cancelled)
	    {
	      gchar *locale = setlocale (LC_CTYPE, "C");
	      sscanf (p, "%lf", &value);
	      setlocale (LC_CTYPE, locale);
	      value = cpse_model_convert (value, units, UNIT_DOTS);
	    }
	}
    }
  while (!(cancelled || in_range (value, range, epsilon)));

  if (value < range->min) value = range->min;
  if (value > range->max) value = range->max;

  if (!cancelled) *v = value;
  return (cancelled);
}

static gboolean
read_double (gdouble *d, const CPSEValue *range, const gchar *description,
	     gint units)
{
  gboolean cancelled;
  gdouble  value;

  require (d);

  if (MAX_UNIT == units)	/* resolution */
    cancelled = read_value (&value, range, description, units, FALSE,
			    "%s (min %.0f, max %.0f) [%.0f %s]");
  else				/* width/height */
    cancelled = read_value (&value, range, description, units, FALSE,
			    "%s (min %.2f, max %.2f) [%.2f %s]");

  if (!cancelled) *d = value;
  return (cancelled);
}

static gboolean
read_integer (gint *i, const CPSEValue *range, const gchar *description)
{
  gboolean cancelled;
  gdouble  value;

  require (i);

  cancelled = read_value (&value, range, description, MAX_UNIT, TRUE,
			  "(%.0f-%.0f) %s");

  if (!cancelled) *i = value;
  return (cancelled);
}

static int
select_ppdfile (CPSEModel *model, gboolean confirm)
{
  gint result = 0;
  gboolean done = FALSE;

  do
    {
      gchar *p;

      printf (" %s%s", _("Filename"), prompt);
      if (-1 == getline (&line_buffer, &line_length, stdin))
	{
	  return (0);
	}
      p = skip_space (chomp (line_buffer));
      done = (0 == strlen (p));
      if (!done)
	{
	  p = strdup (p);
	  require (p);
	  if (ppd_open (p, model->printer.name, model))
	    {
	      result = 1;
	    }
	  else
	    {
	      run_dialog (MSG_TYPE_OKAY, MSG_BAD_FILE);
	      g_free (p);
	      result = 0;
	    }
	  done = (1 == result);
	}
    }
  while (!done);

  return (result);
}

static gint
select_printer (CPSEModel *model, gboolean confirm)
{
  gboolean cancelled = FALSE;
  gboolean selected  = FALSE;
  gint result = 0;

  if (0 == model->dest_sz)
    {
      run_dialog (MSG_TYPE_OKAY, MSG_NO_PRINTERS_FOUND);
      return (0);
    }

  do				/* FIXME: refactor to use read_integer */
    {
      const gchar *p;

      display_printer_menu (model);
      if (-1 == getline (&line_buffer, &line_length, stdin))
	{
	  return (0);
	}
      p = skip_space (line_buffer);
      if (p - line_buffer < line_length)
	{
	  gint index;

	  gchar *locale = setlocale (LC_CTYPE, "C");
	  cancelled = (isalpha (*p) && 'C' == toupper (*p));
	  setlocale (LC_CTYPE, locale);

	  if (!cancelled)
	    {
	      locale = setlocale (LC_CTYPE, "C");
	      selected = ((1 == sscanf (p, "%d", &index))
			  && (0 < index && index <= model->dest_sz));
	      setlocale (LC_CTYPE, locale);

	      if (selected)
		{
		  --index;
		  if (ppd_open (model->printer.ppd_file,
				model->dest[index].name, model))
		    {
		      result = 1;
		    }
		  else
		    {
		      result = 0;
		    }
		}
	    }
	}
    }
  while (!(cancelled || selected));

  if (!cancelled)
    {
      gint reply = run_dialog (MSG_TYPE_YES_NO, MSG_LOAD_PPD_FILE);

      if (MSG_RESPONSE_YES == reply)
	{
	  model->needs_update = (1 == select_ppdfile (model, FALSE));
	}
    }
  return (result);
}

static void
select_units (CPSEModel *m)
{
  static gchar *choices = NULL;

  gboolean done = FALSE;

  require (m);

  if (!choices)			/* lazy initialisation */
    {
      choices = get_menu_line (unit);
      if (!choices)
	{
	  log_errno (ENOMEM);
	  exit (EXIT_FAILURE);
	}
    }

  while (!done)
    {
      action a = ERROR;

      printf ("%s%s", choices, prompt);

      if (-1 == getline (&line_buffer, &line_length, stdin)
	  || 0 == strlen (chomp (line_buffer)))
	return;

      a = parse (unit, line_buffer, line_length);
      if (ERROR != a)
	{
	  m->units = a;
	  done = TRUE;
	}
    }
}

static void
action_save (CPSEModel *m)
{
  gint reply;

  require (m);

  if (!m->needs_update) return;		/* nothing to do */

  if (!m->printer.ppd_file)
    {
      run_dialog (MSG_TYPE_OKAY, MSG_NO_PPDFILE_SELECTED);
      return;
    }

  reply = run_dialog (MSG_TYPE_YES_NO, MSG_REALLY_SAVE);
  if (MSG_RESPONSE_NO == reply)
    return;

  if (ppd_save (m->printer.ppd_file, m))
    {
      if (!m->printer.name)
	m->needs_update = FALSE;
    }
  else
    run_dialog (MSG_TYPE_OKAY, MSG_SAVE_FAILURE);
}

static void
action_apply (CPSEModel *m)
{
  const gchar *file;
  gint reply;

  require (m);

  if (!m->needs_update) return;		/* nothing to do */

  if (!m->printer.name)
    {
      run_dialog (MSG_TYPE_OKAY, MSG_NO_PRINTER_SELECTED);
      return;
    }

  reply = run_dialog (MSG_TYPE_YES_NO, (m->printer.ppd_file
					? MSG_REALLY_SAVE_AND_APPLY
					: MSG_REALLY_APPLY));
  if (MSG_RESPONSE_NO == reply)
    return;

  file = (m->printer.ppd_file
	  ? m->printer.ppd_file : m->printer.cups_ppd);
  if (ppd_save (file, m)
      && ppd_apply (file, m->printer.name))
    {
      m->needs_update = FALSE;
      run_dialog (MSG_TYPE_OKAY, MSG_APPLY_SUCCESS);
    }
  else
    run_dialog (MSG_TYPE_OKAY, MSG_APPLY_FAILURE);
}

static gboolean
get_custom_paper_data (CPSEPaper *p, const CPSEModel *m)
{
  gboolean cancelled = FALSE;
  gboolean selected  = FALSE;
  gchar text[PPD_MAX_TEXT];

  require (p);
  require (m);

  strncpy (text, p->text, PPD_MAX_TEXT);

  p->text[0] = '\0';
  do
    {
      gchar *s;

      if (0 == strlen (text))
	printf (" %s%s", _("Name"), prompt);
      else
	printf (" %s [%s]%s", _("Name"), text, prompt);

      if (-1 == getline (&line_buffer, &line_length, stdin))
	return (cancelled = TRUE);

      s = skip_space (chomp (line_buffer));
      if (0 == strlen (s))
	{
	  cancelled = (0 == strlen (text));
	  s = text;
	}
      if (!cancelled)
	{
	  if (PPD_MAX_TEXT <= strlen (s))
	    s[PPD_MAX_TEXT-1] = '\0';
	  if (s == text
	      || (ppd_is_valid_text (s)
		  && !cpse_model_is_duplicate (m, s)))
	    {
	      strcpy (p->text, s);
	      selected = TRUE;
	    }
	  else
	    {
	      run_dialog (MSG_TYPE_OKAY, MSG_NAME_EXISTS);
	    }
	}
    }
  while (!(cancelled || selected));

  if (!cancelled)
    {
      cancelled = (cancelled
		   || read_double (&p->width, &m->width,
				   _("Width"), m->units));
      cancelled = (cancelled
		   || read_double (&p->height, &m->height,
				   _("Height"), m->units));
      if (m->has_hw_res)
	cancelled = (cancelled
		     || read_double (&p->resolution, &m->resolution,
				     _("Resolution"), MAX_UNIT));

      if (!cancelled)
	{
	  const gchar *units = _(find_units_string (m->units));
	  gint reply;

	  printf ("\n%s: %s\n", _("Name"), p->text);
	  printf ("W: %8.2f %s  H: %8.2f %s",
		  cpse_model_convert (p->width , UNIT_DOTS, m->units), units,
		  cpse_model_convert (p->height, UNIT_DOTS, m->units), units);
	  if (m->has_hw_res)
	    printf ("  R: %4.0f %s", p->resolution,
		    _(find_units_string (MAX_UNIT)));
	  printf ("\n\n");

	  reply = run_dialog (MSG_TYPE_YES_NO, MSG_REGISTER_ENTRY);

	  return (MSG_RESPONSE_YES == reply);
	}
    }
  return (cancelled);
}

static int
index_cmp (const void *a, const void *b)
{
  gint _a = *((gint *) a);
  gint _b = *((gint *) b);

  if (_a == _b) return (0);
  if (_a >  _b) return (1);
  return (-1);
}

static void
custom_register (CPSEModel *m, CPSEPaper *p, size_t index)
{
  require (m);
  require (p);
  require (0 <= index && index < m->custom_paper_sz);

  if (0 != strlen (p->name))
    {
      m->width.val = p->width;
      m->height.val = p->height;
      m->resolution.val = p->resolution;
    }

  if (get_custom_paper_data (p, m))
    {
      if (0 == strlen (p->name))
	{
	  size_t indices[MAX_CUSTOM_PAPERS];
	  gint i = 0;
	  gint c = 0;
	  gint j = 0;

	  /* collect and sort indices in current use */
	  for (i = 0; i < cpse_model_paper_count (m); ++i)
	    sscanf (m->custom_paper[i].name, CUSTOM_PAPER_FORMAT,
		    &indices[i]);
	  qsort (indices, cpse_model_paper_count (m), sizeof (size_t),
		 index_cmp);

	  do			/* find smallest unused index */
	    {
	      i = c + 1;
	      c = indices[j];
	      ++j;
	    }
	  while (MAX_CUSTOM_PAPERS > j && i == c);

	  if (MAX_CUSTOM_PAPERS == j && i == c)
	    {
	      log_err ("internal inconsistency");
	      ++i;
	    }

	  snprintf (p->name, PPD_MAX_NAME, CUSTOM_PAPER_FORMAT, i);
	}
      memcpy (m->custom_paper + index, p, sizeof (*p));
      m->needs_update = TRUE;
    }
}

static void
custom_add (CPSEModel *m)
{
  CPSEPaper custom;
  size_t    index;

  require (m);

  index = cpse_model_paper_count (m);
  if (MAX_CUSTOM_PAPERS <= index)
    {
      run_dialog (MSG_TYPE_OKAY, MSG_TOO_MANY_NAMES);
      return;
    }

  memset (&custom, 0, sizeof (custom));
  custom_register (m, &custom, index);
}

static void
custom_change (CPSEModel *m, gboolean remove)
{
  CPSEPaper custom;
  CPSEValue range;
  gint index;

  require (m);

  range.min = 1;
  range.max = cpse_model_paper_count (m);
  range.val = -1;

  if (range.min > range.max
      || read_integer (&index, &range, _("Custom paper")))
    return;

  --index;
  if (remove)
    {
      gint reply = run_dialog (MSG_TYPE_YES_NO, MSG_REALLY_DELETE);

      if (MSG_RESPONSE_NO == reply)
	return;

      for (; index < range.max - 1; ++index)
	{
	  memcpy (m->custom_paper + index,
		  m->custom_paper + index + 1,
		  sizeof (CPSEPaper));
	}
      memset (m->custom_paper + index, 0, sizeof (CPSEPaper));
      m->needs_update = TRUE;
    }
  else
    {
      memcpy (&custom, m->custom_paper + index, sizeof (custom));
      custom_register (m, &custom, index);
    }
}

int
main (int argc, char *argv[])
{
  CPSEModel *model = NULL;

  size_t level = 1;

  setvbuf (stdout, NULL, _IONBF, 0);

  /* Sorry, but we haven't got around to all the finicky formatting
     details to get everything to display nice on a console in your
     preferred language.  Patches are welcome ;-)
   */
  setlocale (LC_ALL, "C");

  if (1 < argc)
    log_errno (ENOTSUP);

  model = cpse_model_new ();
  if (!model)
    {
      log_errno (ENOMEM);
      return (EXIT_FAILURE);
    }

  while (0 < level)
    {
      action a = get_action (level, model);

      switch (a)
	{
	case ERROR:
	  break;
	case SELECT_PPDFILE:
	  level += select_ppdfile (model, TRUE);
	  break;
	case SELECT_PRINTER:
	  level += select_printer (model, TRUE);
	  break;
	case SELECT_UNITS:
	  select_units (model);
	  break;
	case ACTION_EDIT:
	  ++level;
	  break;
	case ACTION_SAVE:
	  action_save (model);
	  break;
	case ACTION_APPLY:
	  action_apply (model);
	  break;
	case CUSTOM_ADD:
	  custom_add (model);
	  break;
	case CUSTOM_CHANGE:
	  custom_change (model, FALSE);
	  break;
	case CUSTOM_REMOVE:
	  custom_change (model, TRUE);
	  break;
	case QUIT:
	  --level;
	  if (1 == level && model->needs_update)
	    {
	      gint reply = run_dialog (MSG_TYPE_YES_NO, MSG_REALLY_QUIT);

	      if (MSG_RESPONSE_YES == reply)
		cpse_model_reset (model);
	      else
		++level;
	    }
	  break;
	default:
	  log_msg ("unknown action: %d", a);
	}
    }

  return (EXIT_SUCCESS);
}
