/*  cpse-gtk.c -- GTK+ based GUI for CPSE
    Copyright (C) 2008  SEIKO EPSON CORPORATION

    This file is part of "Custom Paper Size Editor".

    "Custom Paper Size Editor" is free software.
    You can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License or at your option any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY;  without even the implied warranty of FITNESS
    FOR A PARTICULAR PURPOSE or MERCHANTABILITY.
    See the GNU General Public License for more details.

    You should have received a verbatim copy of the GNU General Public
    License along with this program; if not, write to:

        Free Software Foundation, Inc.
        59 Temple Place, Suite 330
        Boston, MA  02111-1307  USA
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <locale.h>
#include <stdlib.h>
#include <string.h>

#include "generix.h"
#include "message.h"
#include "model.h"
#include "ppd.h"


#define CPSE_GLADE	G_DIR_SEPARATOR_S "cpse.glade"
#define CPSE_LOGFILE    "/dev/null"

#define CUPS_PPD_DIR	"/usr/share/cups/model"
#define LSB_PPD_DIR	"/usr/share/ppd"
#define DEFAULT_PPD_DIR	CUPS_PPD_DIR

static const gchar *const default_printer_lbl =
  N_("<b>Printer / PPD File</b>");

static const gchar *const ppd_dir[] = {
  DEFAULT_PPD_DIR,		/* has to be first */
  LSB_PPD_DIR,
  CUPS_PPD_DIR,
  NULL				/* terminator */
};

static const gchar *const unit_name[] = {
  N_("cm"),
  N_("inch"),
#if ENABLE_EXTRA_UNITS
  N_("mm"),
  N_("dots"),
#endif
  NULL				/* terminator */
};

#define CALLBACK
#define check_p	require


GladeXML  *xml	 = NULL;
CPSEModel *model = NULL;

#if HAVE_GTK_COMBO_BOX_GET_ACTIVE_TEXT

#define my_combo_box_get_active_text	gtk_combo_box_get_active_text

#else  /* !HAVE_GTK_COMBO_GET_ACTIVE_TEXT */

static gchar *
my_combo_box_get_active_text (GtkComboBox *b)
{
  gchar *result = NULL;

  GtkTreeModel *m = gtk_combo_box_get_model (b);
  GtkTreeIter   it;

  require (m);

  if (gtk_combo_box_get_active_iter (b, &it))
    gtk_tree_model_get (m, &it, 0, &result, -1);

  return (result);
}

#endif /* !HAVE_GTK_COMBO_GET_ACTIVE_TEXT */


/*! Converts a length in dots to a length in the GUI's selected units.
    Note that the active text we get from the combobox is likely to be
    internationalised.	Hence the wrapping in _() before comparison of
    the strings.
 */
static gdouble
convert (gdouble dots)
{
  GtkComboBox *widget;
  gchar  *units;
  gdouble result = dots;

  widget = GTK_COMBO_BOX (glade_xml_get_widget (xml, "units"));
  check_p (widget);

  units = my_combo_box_get_active_text (widget);
  check_p (units);

  if (0 == strcmp (units, _("cm")))
    result = (dots / DOTS_PER_INCH * CM_PER_INCH);
  else if (0 == strcmp (units, _("mm")))
    result = (dots / DOTS_PER_INCH * MM_PER_INCH);
  else if (0 == strcmp (units, _("inch")))
    result = (dots / DOTS_PER_INCH);
  else if (0 == strcmp (units, _("dots")))
    result = (dots);
  else
    log_msg ("Request to convert to unsupported unit type: %s", units);

  g_free (units);

  return (result);
}

/*! Converts a length in the GUI's selected units to a length in dots.
    The function name is 'convert' spelled backwards, indicating that
    it does the reverse of that function.
    Note that the active text we get from the combobox is likely to be
    internationalised.	Hence the wrapping in _() before comparison of
    the strings.
    Note also that repeated back-and-forth conversion may lead to loss
    of precision.
 */
static gdouble
trevnoc (gdouble value)
{
  GtkComboBox *widget;
  gchar  *units;
  gdouble result = value;

  widget = GTK_COMBO_BOX (glade_xml_get_widget (xml, "units"));
  check_p (widget);

  units = my_combo_box_get_active_text (widget);
  check_p (units);

  if (0 == strcmp (units, _("cm")))
    result = (value / CM_PER_INCH * DOTS_PER_INCH);
  else if (0 == strcmp (units, _("mm")))
    result = (value / MM_PER_INCH * DOTS_PER_INCH);
  else if (0 == strcmp (units, _("inch")))
    result = (value * DOTS_PER_INCH);
  else if (0 == strcmp (units, _("dots")))
    result = (value);
  else
    log_msg ("Request to convert to unsupported unit type: %s", units);

  g_free (units);

  return (result);
}

static void
set_limit (const gchar *name, gdouble value)
{
  GtkEntry    *widget;
  GtkComboBox *units;

  gchar *format = "%.2f %s";
  gchar *units_text;

  widget = GTK_ENTRY (glade_xml_get_widget (xml, name));
  check_p (widget);
  units = GTK_COMBO_BOX (glade_xml_get_widget (xml, "units"));
  check_p (units);

  units_text = my_combo_box_get_active_text (units);
  format = g_strdup_printf (format, convert (value), units_text);
  g_free (units_text);

  gtk_entry_set_text (widget, format);
  g_free (format);
}

static void
zap_limit (const gchar *name)
{
  GtkEntry *widget;

  widget = GTK_ENTRY (glade_xml_get_widget (xml, name));
  check_p (widget);

  gtk_entry_set_text (widget, "");
}

static void
set_sensitive (const gchar *name, gboolean sensitive)
{
  GtkWidget *widget;

  widget = glade_xml_get_widget (xml, name);
  check_p (widget);

  gtk_widget_set_sensitive (widget, sensitive);
}

static void
reset_spin_button (const gchar *name, CPSEValue value)
{
  GtkSpinButton *widget;

  widget = GTK_SPIN_BUTTON (glade_xml_get_widget (xml, name));
  check_p (widget);

  if (0 != strcmp ("resolution", name))
    {
      value.min = convert (value.min);
      value.max = convert (value.max);
      value.val = convert (value.val);
    }
  gtk_spin_button_set_range (widget, value.min, value.max);
  gtk_spin_button_set_value (widget, value.val);
}

static void
zap_spin_button (const gchar *name)
{
  GtkSpinButton *widget;

  widget = GTK_SPIN_BUTTON (glade_xml_get_widget (xml, name));
  check_p (widget);

  gtk_spin_button_set_range (widget, 0, 0);
  gtk_spin_button_set_value (widget, 0);
}

static void
set_visibility (const gchar *name, gboolean show)
{
  GtkWidget *widget;

  widget = GTK_WIDGET (glade_xml_get_widget (xml, name));
  check_p (widget);

  if (show)
    gtk_widget_show (widget);
  else
    gtk_widget_hide (widget);
}

/*! Updates the list of custom page size definitions in the model.

    This is necessary because the GTK app opts to maintain the custom
    paper size info in a GtkTreeModel between loading data from file
    (or printer) and saving/applying it.  This was deemed easier than
    keeping the underlying \c model in sync with every little change
    in the GUI.
 */
static void
update_model (void)
{
  GtkTreeModel *tree;
  GtkTreeIter   it;
  GtkTreeView  *w;

  gint i = 0;

  w = GTK_TREE_VIEW (glade_xml_get_widget (xml, "custom_pager"));
  check_p (w);

  tree = GTK_TREE_MODEL (gtk_tree_view_get_model (w));
  check_p (tree);

  if (gtk_tree_model_get_iter_first (tree, &it))
    {					/* have tree model data */
      do
	{
	  CPSEPaper *p = model->custom_paper + i;
	  gchar *name = NULL;
	  gchar *text = NULL;

	  gtk_tree_model_get (tree, &it,
			      COL_TEXT, &text,
			      COL_WIDTH, &(p->width),
			      COL_HEIGHT, &(p->height),
			      COL_RESOLUTION, &(p->resolution),
			      COL_NAME, &name,
			      -1);

	  g_strlcpy (p->name, name, PPD_MAX_NAME);
	  g_strlcpy (p->text, text, PPD_MAX_TEXT);

	  g_free (name);
	  g_free (text);

	  ++i;
	}
      while (gtk_tree_model_iter_next (tree, &it)
	     && i <= MAX_CUSTOM_PAPERS);
    }
  for (;i < MAX_CUSTOM_PAPERS; ++i)	/* clear remainder */
    {
      memset (model->custom_paper + i, '\0', sizeof (CPSEPaper));
    }
}

CALLBACK void
update_value (GtkSpinButton *widget, gpointer data)
{
  const gchar *name;
  gdouble      value;

  check_p (widget);

  name = gtk_widget_get_name (GTK_WIDGET (widget));
  value = gtk_spin_button_get_value (widget);

  if (0 == strcmp ("width", name))
    model->width.val  = trevnoc (value);
  else if (0 == strcmp ("height", name))
    model->height.val = trevnoc (value);
  else if (0 == strcmp ("resolution", name))
    model->resolution.val = value;
}

CALLBACK void
change_limit_units (GtkWidget *widget, gpointer data)
{
  const gchar *name;

  check_p (widget);

  name = gtk_widget_get_name (widget);

  if (0 == strcmp ("width.min", name))
    set_limit (name, model->width.min);
  else if (0 == strcmp ("height.min", name))
    set_limit (name, model->height.min);
  else if (0 == strcmp ("width.max", name))
    set_limit (name, model->width.max);
  else if (0 == strcmp ("height.max", name))
    set_limit (name, model->height.max);
}

CALLBACK void
change_spin_units (GtkWidget *widget, gpointer data)
{
  const gchar *name;

  check_p (widget);

  name = gtk_widget_get_name (widget);

  if (0 == strcmp ("width", name))
    reset_spin_button (name, model->width);
  else if (0 == strcmp ("height", name))
    reset_spin_button (name, model->height);
}

CALLBACK void
change_editor_units (GtkTreeView *widget, gpointer data)
{
  check_p (widget);
  gtk_tree_view_columns_autosize (widget);
}

CALLBACK gboolean
select_file (GtkEntry *widget, GdkEventButton *event, gpointer ppd_file);

CALLBACK void
select_printer (GtkEntry *widget, gpointer data);

static void
set_default_directory (void)
{
  GtkEntry *widget;
  gboolean success      = FALSE;
  const gchar *const *p = ppd_dir;

  widget = GTK_ENTRY (glade_xml_get_widget (xml, "file_path_editor"));
  check_p (widget);

  while (!success && p && *p)
    {
      if (g_file_test (*p, G_FILE_TEST_IS_DIR))
	{
	  success = TRUE;
	  gtk_entry_set_text (widget, *p);
	}
      ++p;
    }
}

static void
reset_printer_name (gboolean block)
{
  GtkComboBox  *w;
  GtkTreeModel *tree;
  GtkTreeIter   it;

  w = GTK_COMBO_BOX (glade_xml_get_widget (xml, "printer_chooser"));
  require (w);
  tree = gtk_combo_box_get_model (w);
  require (tree);

  if (!gtk_tree_model_get_iter_first (tree, &it))
    return;

  if (model->printer.name && 0 != strlen (model->printer.name))
    {
      gboolean found = FALSE;
      do
	{
	  gchar *text = NULL;

	  gtk_tree_model_get (tree, &it, COL_TEXT, &text, -1);
	  found = (0 == strcmp (text, model->printer.name));

	  g_free (text);
	}
      while (!found && gtk_tree_model_iter_next (tree, &it));

      if (!found)
	log_err ("internal inconsistency");
    }

  if (block)
    g_signal_handlers_block_matched (w, G_SIGNAL_MATCH_FUNC, 0, 0, NULL,
				     select_printer, NULL);
  gtk_combo_box_set_active_iter (w, &it);
  if (block)
    g_signal_handlers_unblock_matched (w, G_SIGNAL_MATCH_FUNC, 0, 0, NULL,
				       select_printer, NULL);
}

/* Note this function is called by select_ppd_file(), select_file()
   and select_printer().  These three callbacks set up the arguments
   to meet our expectations.

   - select_ppd_file() passes a NULL data pointer
   - select_file() passes a pointer to a CPSEPrinter object with its
     ppd_file member set
   - select_printer() passes a pointer to a CPSEPrinter object with
     its printer member set

   We just copy "upstream" data into our own CPSEPrinter object and
   fill out the rest based on current settings where necessary.
 */
static gboolean
_select_ppd_file (GtkEntry *widget, GdkEventButton *event, gpointer data)
{
  CPSEPrinter printer;

  const gchar *fake_ppd_fmt = "<%s.ppd>";
        gchar *fake_ppd     = NULL;

  gboolean unselecting_printer = FALSE;

  memset (&printer, '\0', sizeof (printer));
  if (data)
    {
      CPSEPrinter *p = (CPSEPrinter *) data;
      printer.model = p->model;
      printer.name  = p->name;
      printer.ppd_file = p->ppd_file;
      printer.cups_ppd = p->cups_ppd;
    }

  if (!printer.ppd_file && !printer.name)
    printer.ppd_file = gtk_entry_get_text (widget);

  if (!printer.name)
    {
      GtkComboBox *w;
      w = GTK_COMBO_BOX (glade_xml_get_widget (xml, "printer_chooser"));
      check_p (w);
      printer.name = my_combo_box_get_active_text (w);
    }
  else
    unselecting_printer = (0 == strlen (printer.name));

  /* We don't want to pop up the file selection dialog when a "fake"
     PPD file is displayed.
   */
  if (printer.ppd_file)
    {
      fake_ppd = g_strdup_printf (fake_ppd_fmt, printer.name);
      if (0 == strcmp (fake_ppd, printer.ppd_file))
	printer.ppd_file = NULL;
      g_free (fake_ppd);
      fake_ppd = NULL;
    }

  if (printer.ppd_file
      && !g_file_test (printer.ppd_file, G_FILE_TEST_IS_REGULAR))
    {
      /* Just open up a GtkFileChooser for now but ideally we would
	 notify the user if they had entered a non-existent file.  I
	 don't think we should show a notification if they entered a
	 directory name.
       */
      return (select_file (widget, NULL, NULL));
    }

  if (model->needs_update)
    {
      gint reply;

      reply = run_dialog (MSG_TYPE_YES_NO, MSG_CLOBBER_SETTINGS);
      if (GTK_RESPONSE_NO == reply)
	{
	  reset_printer_name (TRUE);
	  return (TRUE);
	}
    }

  if (unselecting_printer)
    printer.ppd_file = model->printer.ppd_file;

  if (!ppd_open (printer.ppd_file, printer.name, model))
    {
      if (!printer.ppd_file)
	reset_printer_name (TRUE);
      run_dialog (MSG_TYPE_OKAY, MSG_BAD_FILE);
      return (FALSE);
    }

  /* The PPD file or printer has been changed successfully, now update
     the view. */
  {
    GtkLabel *w;

    w = GTK_LABEL (glade_xml_get_widget (xml, "printer_lbl"));
    check_p (w);
    if (model->printer.model)
      gtk_label_set_markup (w, g_strdup_printf ("<b>%s</b>", /* leak */
						model->printer.model));
    else
      gtk_label_set_markup (w, _(default_printer_lbl));
  }
  {
    if (model->printer.cups_ppd)
      {
	fake_ppd = g_strdup_printf (fake_ppd_fmt, model->printer.name);

	gtk_entry_set_text (widget, fake_ppd);
	g_free (fake_ppd);
	fake_ppd = NULL;
      }
    else if (model->printer.ppd_file)
      {
	gtk_entry_set_text (widget, model->printer.ppd_file);
      }
    else
      {
	set_default_directory ();
      }
  }


  {				/* clear custom paper name */
    GtkEntry *w;

    w = GTK_ENTRY (glade_xml_get_widget (xml, "name"));
    check_p (w);

    gtk_entry_set_text (w, "");
  }
  {				/* update the editor's model */
    GtkListStore *list;
    GtkTreeIter   it;
    GtkTreeView  *w;

    gint i;

    w = GTK_TREE_VIEW (glade_xml_get_widget (xml, "custom_pager"));
    check_p (w);

    list = GTK_LIST_STORE (gtk_tree_view_get_model (w));
    check_p (list);

    gtk_list_store_clear (list);
    for (i = 0; i < MAX_CUSTOM_PAPERS; ++i)
      {
	const CPSEPaper *p = model->custom_paper + i;

	if (0 == strlen (p->name)) continue;

	gtk_list_store_append (list, &it);
	gtk_list_store_set (list, &it,
			    COL_TEXT, p->text,
			    COL_WIDTH, p->width,
			    COL_HEIGHT, p->height,
			    COL_RESOLUTION, p->resolution,
			    COL_NAME, p->name,
			    -1);
      }
    gtk_tree_view_column_set_visible
      (gtk_tree_view_get_column (w, COL_RESOLUTION), model->has_hw_res);
  }

  if (model->printer.ppd_file || model->printer.cups_ppd)
    {
				/* update limits */
      set_limit ("width.min" , model->width.min);
      set_limit ("height.min", model->height.min);
      set_limit ("width.max" , model->width.max);
      set_limit ("height.max", model->height.max);

				/* update spin buttons */
      reset_spin_button ("width" , model->width);
      reset_spin_button ("height", model->height);
      reset_spin_button ("resolution", model->resolution);

      set_visibility ("res_lbl", model->has_hw_res);
      set_visibility ("resolution", model->has_hw_res);

      set_sensitive ("editor", TRUE);
      set_sensitive ("units" , TRUE);
      set_sensitive ("save" , (NULL != model->printer.ppd_file));
      set_sensitive ("apply", (NULL != model->printer.name
			       && 0 < strlen (model->printer.name)));
    }
  else
    {
      zap_limit ("width.min");
      zap_limit ("height.min");
      zap_limit ("width.max");
      zap_limit ("height.max");

      zap_spin_button ("width");
      zap_spin_button ("height");
      zap_spin_button ("resolution");

      set_visibility ("res_lbl", FALSE);
      set_visibility ("resolution", FALSE);

      set_sensitive ("editor", FALSE);
      set_sensitive ("units" , FALSE);
      set_sensitive ("save" , FALSE);
      set_sensitive ("apply", FALSE);
    }

  return (FALSE);
}

/* Make sure the real _select_ppd_file () gets called with correctly
   initialised arguments.
 */
CALLBACK gboolean
select_ppd_file (GtkEntry *widget, GdkEventButton *event, gpointer data)
{
  return (_select_ppd_file (widget, event, NULL));
}

/*! Sets up and displays a file selection dialog for the user to
    select a PPD file candidate.  It starts the dialog with the
    longest path that points to an existing directory.  The path
    is gotten from the entry box.

    Note, this function should \e not modify any of the model's
    members.
 */
CALLBACK gboolean
select_file (GtkEntry *widget, GdkEventButton *event, gpointer data)
{
  GtkFileChooser *dialog;

  gchar *path;
  gint	 reply;

  CPSEPrinter printer;
  memset (&printer, '\0', sizeof (printer));

  check_p (widget);

  dialog = GTK_FILE_CHOOSER (glade_xml_get_widget (xml, "file_browser"));
  check_p (dialog);

  path = g_strdup (gtk_entry_get_text (widget));
  while (path
	 && 0 != strcmp (G_DIR_SEPARATOR_S, path)
	 && 0 != strcmp (".", path)
	 && !g_file_test (path, G_FILE_TEST_IS_DIR))
    {
      gchar *path_dirname = g_path_get_dirname (path);
      g_free (path);
      path = path_dirname;
    }
  if (!path)
    path = g_strdup (g_get_home_dir ());

  if (0 == strcmp (".", path))
    {
      g_free (path);
      path = g_get_current_dir ();
    }

  gtk_file_chooser_set_current_folder (dialog, path);
  g_free (path);		/* path no longer needed */

  reply = gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_hide (GTK_WIDGET (dialog));

  if (GTK_RESPONSE_ACCEPT != reply)
    return (FALSE);

  printer.ppd_file = gtk_file_chooser_get_filename (dialog); /* possible leak */

  if (!printer.ppd_file) return (FALSE);

  return (_select_ppd_file (widget, NULL, (gpointer) &printer));
}

CALLBACK void
select_printer (GtkEntry *widget, gpointer data)
{
  CPSEPrinter printer;

  check_p (data);
  memset (&printer, '\0', sizeof (printer));
  printer.name = my_combo_box_get_active_text (data); /* possible leak */
  _select_ppd_file (widget, NULL, (gpointer) &printer);
}

CALLBACK void
select_entry (GtkTreeView *widget, gpointer data)
{
  GtkTreeSelection *selection;
  GtkTreeModel	   *list;
  GtkTreeIter	    it;

  check_p (widget);

  selection = gtk_tree_view_get_selection (widget);
  if (selection
      && gtk_tree_selection_get_selected (selection, &list, &it))
    {
      GtkEntry	    *name;
      GtkSpinButton *width;
      GtkSpinButton *height;
      GtkSpinButton *resolution;
      gchar	    *text = NULL;

      name = GTK_ENTRY (glade_xml_get_widget (xml, "name"));
      check_p (name);
      width = GTK_SPIN_BUTTON (glade_xml_get_widget (xml, "width"));
      check_p (width);
      height = GTK_SPIN_BUTTON (glade_xml_get_widget (xml, "height"));
      check_p (height);
      resolution = GTK_SPIN_BUTTON (glade_xml_get_widget (xml, "resolution"));
      check_p (resolution);

      gtk_tree_model_get (list, &it,
			  COL_TEXT	, &text,
			  COL_WIDTH	, &(model->width.val),
			  COL_HEIGHT	, &(model->height.val),
			  COL_RESOLUTION, &(model->resolution.val),
			  /* we don't need the COL_NAME value */
			  -1);

      gtk_entry_set_text (name, text);
      gtk_spin_button_set_value (width , convert (model->width.val));
      gtk_spin_button_set_value (height, convert (model->height.val));
      gtk_spin_button_set_value (resolution, model->resolution.val);
      g_free (text);
    }
}

static gboolean
tree_collect_ids (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *it,
		  gpointer data)
{
  gchar *id = NULL;
  gint   n = -1;

  gtk_tree_model_get (model, it, COL_NAME, &id, -1);
  if (id && 1 == sscanf (id, CUSTOM_PAPER_FORMAT, &n) && 0 < n)
    {
      g_array_append_val ((GArray *) data, n);
    }
  g_free (id);

  return (FALSE);
}

static gint
id_cmp (gconstpointer a, gconstpointer b)
{
  gint _a = *((gint *) a);
  gint _b = *((gint *) b);

  if (_a == _b) return (0);
  if (_a >  _b) return (1);
  return (-1);
}

/* Caller need to g_free() the string returned.
 */
static gchar *
tree_create_id (GtkTreeModel *tree)
{				/* create a new, unused ID */
  GArray *a;

  gint i = 0;
  gint j = 0;
  gint c = 0;

  if (!tree) return (NULL);

  a = g_array_new (TRUE, TRUE, sizeof (gint));
  if (!a) return (NULL);

  gtk_tree_model_foreach (tree, tree_collect_ids, a);
  g_array_sort (a, id_cmp);

  do				/* find smallest unused index */
    {
      i = c + 1;
      c = g_array_index (a, gint, j);
      ++j;
    }
  while (MAX_CUSTOM_PAPERS > j && i == c);

  if (MAX_CUSTOM_PAPERS == j && i == c)
    {
      log_err ("internal inconsistency");
      ++i;
    }
  g_array_free (a, TRUE);

  return (g_strdup_printf (CUSTOM_PAPER_FORMAT, i));
}

static gboolean
tree_iter_count (GtkTreeModel *model, GtkTreePath *path,
		 GtkTreeIter *it, gpointer data)
{
  ++(*((gint *) (data)));
  return (FALSE);
}

static gboolean
tree_max_reached (GtkTreeModel *tree)
{
  gint i = 0;

  if (!tree) return (FALSE);

  gtk_tree_model_foreach (tree, tree_iter_count, &i);

  return (MAX_CUSTOM_PAPERS == i);
}

static gboolean
tree_is_duplicate (GtkTreeModel *tree, const gchar *text)
{
  gboolean found = FALSE;
  GtkTreeIter it;

  require (tree);
  require (text);

  if (gtk_tree_model_get_iter_first (tree, &it))
    {
      do
	{
	  gchar *t = NULL;

	  gtk_tree_model_get (tree, &it, COL_TEXT, &t, -1);
	  found = (0 == strcmp (t, text));

	  g_free (t);
	}
      while (!found && gtk_tree_model_iter_next (tree, &it));
    }
  return (found);
}

CALLBACK gboolean
add_entry (GtkTreeView *widget, GdkEventButton *event, gpointer data)
{
  GtkListStore *list;
  GtkEntry     *name;
  GtkTreeIter	it;

  gchar *text;
  gchar *id_name = NULL;

  check_p (widget);

  list = GTK_LIST_STORE (gtk_tree_view_get_model (widget));
  check_p (list);
  name = GTK_ENTRY (glade_xml_get_widget (xml, "name"));
  check_p (name);

  if (tree_max_reached (GTK_TREE_MODEL (list)))
    {
      run_dialog (MSG_TYPE_OKAY, MSG_TOO_MANY_NAMES);
      return (FALSE);
    }

  text = g_strchug (g_strdup (gtk_entry_get_text (name)));
  gtk_entry_set_text (name, text);
  if (0 == strlen (text))
    {
      g_free (text);
      run_dialog (MSG_TYPE_OKAY, MSG_MISSING_ITEMS);
      return (FALSE);
    }

  if (!ppd_is_valid_text (text)
      || tree_is_duplicate (GTK_TREE_MODEL (list), text))
    {
      g_free (text);
      run_dialog (MSG_TYPE_OKAY, MSG_NAME_EXISTS);
      return (FALSE);
    }

  id_name = tree_create_id (GTK_TREE_MODEL (list));

  gtk_list_store_append (list, &it);
  gtk_list_store_set (list, &it,
		      COL_TEXT	    , text,
		      COL_WIDTH	    , model->width.val,
		      COL_HEIGHT    , model->height.val,
		      COL_RESOLUTION, model->resolution.val,
		      COL_NAME      , id_name,
		      -1);
  g_free (text);
  g_free (id_name);

  model->needs_update = TRUE;

  return (FALSE);
}

CALLBACK gboolean
edit_entry (GtkTreeView *widget, GdkEventButton *event, gpointer data)
{
  GtkTreeSelection *selection;
  GtkTreeIter	    it;

  check_p (widget);

  selection = gtk_tree_view_get_selection (widget);
  if (selection
      && gtk_tree_selection_get_selected (selection, NULL, &it))
    {
      GtkListStore *list;
      GtkEntry	   *name;
      gchar	   *text;
      gchar	   *current = NULL;

      list = GTK_LIST_STORE (gtk_tree_view_get_model (widget));
      check_p (list);
      name = GTK_ENTRY (glade_xml_get_widget (xml, "name"));
      check_p (name);

      text = g_strchug (g_strdup (gtk_entry_get_text (name)));
      gtk_entry_set_text (name, text);
      if (0 == strlen (text))
	{
	  g_free (text);
	  run_dialog (MSG_TYPE_OKAY, MSG_MISSING_ITEMS);
	  return (FALSE);
	}

      gtk_tree_model_get (GTK_TREE_MODEL (list), &it, COL_TEXT, &current, -1);

      if (0 != strcmp (current, text)
	  && (!ppd_is_valid_text (text)
	      || tree_is_duplicate (GTK_TREE_MODEL (list), text)))
	{
	  g_free (text);
	  g_free (current);
	  run_dialog (MSG_TYPE_OKAY, MSG_NAME_EXISTS);
	  return (FALSE);
	}

      gtk_list_store_set (list, &it,
			  COL_TEXT	, text,
			  COL_WIDTH	, model->width.val,
			  COL_HEIGHT	, model->height.val,
			  COL_RESOLUTION, model->resolution.val,
			  /* the COL_NAME value is a constant */
			  -1);
      g_free (text);
      g_free (current);
      model->needs_update = TRUE;
    }
  else
    {
      run_dialog (MSG_TYPE_OKAY, MSG_NO_SELECTION);
    }

  return (FALSE);
}

CALLBACK gboolean
remove_entry (GtkTreeView *widget, GdkEventButton *event, gpointer data)
{
  GtkTreeSelection *selection;
  GtkTreeIter	    it;

  check_p (widget);

  selection = gtk_tree_view_get_selection (widget);
  if (selection
      && gtk_tree_selection_get_selected (selection, NULL, &it))
    {
      GtkListStore *list;

      list = GTK_LIST_STORE (gtk_tree_view_get_model (widget));
      check_p (list);

      gtk_list_store_remove (list, &it);
      model->needs_update = TRUE;
    }
  else
    {
      run_dialog (MSG_TYPE_OKAY, MSG_NO_SELECTION);
    }

  return (FALSE);
}

CALLBACK void
save_changes (void)
{
  gint reply;

  if (!model->needs_update)
    return;

  reply = run_dialog (MSG_TYPE_YES_NO, MSG_REALLY_SAVE);
  if (GTK_RESPONSE_NO == reply)
    return;

  update_model ();

  if (ppd_save (model->printer.ppd_file, model))
    {
      model->needs_update = FALSE;
    }
  else
    {
      run_dialog (MSG_TYPE_OKAY, MSG_SAVE_FAILURE);
    }
}

CALLBACK void
apply_changes (void)
{
  const gchar *file;
  gint reply;

  if (!model->needs_update && !model->printer.ppd_file)
    return;

  reply = run_dialog (MSG_TYPE_YES_NO, (model->printer.ppd_file
					? MSG_REALLY_SAVE_AND_APPLY
					: MSG_REALLY_APPLY));
  if (GTK_RESPONSE_NO == reply)
    return;

  update_model ();

  file = (model->printer.ppd_file
	  ? model->printer.ppd_file : model->printer.cups_ppd);
  if (ppd_save (file, model)
      && ppd_apply (file, model->printer.name))
    {
      model->needs_update = FALSE;
      run_dialog (MSG_TYPE_OKAY, MSG_APPLY_SUCCESS);
    }
  else
    {
      run_dialog (MSG_TYPE_OKAY, MSG_APPLY_FAILURE);
    }
}

CALLBACK void
main_quit (void)
{
  if (model->needs_update)
    {
      gint reply;

      reply = run_dialog (MSG_TYPE_YES_NO, MSG_REALLY_QUIT);
      if (GTK_RESPONSE_NO == reply)
	return;
    }
  gtk_main_quit ();
}

CALLBACK gboolean
frame_quit (GtkWindow *window, GdkEvent *event, gpointer data)
{
  if (GDK_DELETE == event->type)
    main_quit ();

  return (FALSE);
}


static void 
value_formatter (GtkTreeViewColumn *t_view_c,
		 GtkCellRenderer   *renderer,
		 GtkTreeModel	   *list,
		 GtkTreeIter	   *it,
		 gpointer	   data)
{
  gdouble value;
  gchar *text;
  gint column;

  const gchar *title = (const gchar *) data;
  if (0 == strcmp (_("Width"), title))
    column = COL_WIDTH;
  if (0 == strcmp (_("Height"), title))
    column = COL_HEIGHT;
  if (0 == strcmp (_("Resolution"), title))
    column = COL_RESOLUTION;

  gtk_tree_model_get (list, it, column, &value, -1);
  text = g_strdup_printf (((COL_RESOLUTION == column) ? "%.1f" : "%.2f"),
			  ((COL_RESOLUTION == column)
			   ? value : convert (value)));
  g_object_set (renderer, "text", text, NULL);
  g_free (text);
}

static void
tree_view_insert_column (GtkTreeView *widget,
			 gint column, const gchar *title)
{
  GtkTreeViewColumn *t_view_c = gtk_tree_view_column_new ();
  GtkCellRenderer   *renderer = gtk_cell_renderer_text_new ();

  check_p (widget);
  check_p (title);
  check_p (t_view_c);
  check_p (renderer);

  gtk_tree_view_column_set_title (t_view_c, title);
  gtk_tree_view_column_pack_start (t_view_c, renderer, TRUE);
  gtk_tree_view_column_set_attributes (t_view_c, renderer,
				       "text", column, NULL);

  gtk_tree_view_column_set_alignment (t_view_c, 0.5);
  gtk_tree_view_column_set_resizable (t_view_c, TRUE);
  gtk_tree_view_column_set_expand (t_view_c, FALSE);

  switch (column)
    {
    case COL_TEXT:
      {
	gtk_tree_view_column_set_expand (t_view_c, TRUE);
	gtk_tree_view_column_set_sort_column_id (t_view_c, column);
	break;
      }
    case COL_NAME:
      {
	break;
      }
    case COL_RESOLUTION:
      {
	gtk_tree_view_column_set_visible (t_view_c, model->has_hw_res);
	/* fall through */
      }
    default:
      {
	g_object_set (G_OBJECT (renderer), "xalign", 1.0, NULL);
	gtk_tree_view_column_set_cell_data_func (t_view_c, renderer,
						 value_formatter,
						 (gpointer) title, NULL);
      }
    }
  gtk_tree_view_append_column (widget, t_view_c);
}


int
main (int argc, char *argv[])
{
  if (!getenv ("CPSE_DEVEL"))
    {
      stderr = freopen(CPSE_LOGFILE, "w", stderr);
      if (!stderr)
	{
	  printf("error: can't open %s.\n", CPSE_LOGFILE);
	}
    }
  else
    {
      printf("CPSE_DEVEL mode.\n");
    }

  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE_TARNAME, LOCALEDIR);
  bind_textdomain_codeset (PACKAGE_TARNAME, "UTF-8");
  textdomain (PACKAGE_TARNAME);
  gtk_set_locale ();

  gtk_init (&argc, &argv);

  /* load the interface */
  if (!getenv ("CPSE_DEVEL"))
    xml = glade_xml_new (PKGDATADIR CPSE_GLADE, NULL, NULL);
  else
    xml = glade_xml_new ("." CPSE_GLADE, NULL, NULL);
  check_p (xml);

  /* create the model and connect it to the interface */
  model = cpse_model_new ();
  check_p (model);

  {				/* set default printer label */
    GtkLabel *widget;

    widget = GTK_LABEL (glade_xml_get_widget (xml, "printer_lbl"));
    check_p (widget);

    gtk_label_set_markup (widget, _(default_printer_lbl));
  }
  {				/* add CUPS printers */
    GtkComboBox  *widget;
    GtkListStore *list;
    GtkTreeIter   it;

    gint i;

    widget = GTK_COMBO_BOX (glade_xml_get_widget (xml, "printer_chooser"));
    check_p (widget);

    list = gtk_list_store_new (1, G_TYPE_STRING);
    check_p (list);

    /* the empty string allows for unselecting the printer */
    gtk_list_store_append (list, &it);
    gtk_list_store_set (list, &it, 0, "", -1);

    for (i = 0; i < model->dest_sz; ++i)
      {
	gchar *name = g_strdup (model->dest[i].name);
	if (!name) continue;
	gtk_list_store_append (list, &it);
	gtk_list_store_set (list, &it, 0, name, -1);
	g_free (name);
      }
    gtk_combo_box_set_model (widget, GTK_TREE_MODEL (list));
  }
  set_default_directory ();
  {				/* set units */
    GtkComboBox  *widget;

    const gchar *const *p = unit_name;

    widget = GTK_COMBO_BOX (glade_xml_get_widget (xml, "units"));
    check_p (widget);

    /* Remove the 'stub' forcing Glade to create a text GtkComboBox
       and set our list of unit names.  This is so that translators
       are not bothered with the full list of unit names as a single
       string to translate while still allowing translated units.
     */
    gtk_combo_box_remove_text (widget, 0);
    while (p && *p)
      {
	gtk_combo_box_append_text (widget, _(*p));
	++p;
      }
    gtk_combo_box_set_active (widget, 0);
  }
  {				/* configure the custom pager */
    GtkTreeView *widget;

    widget = GTK_TREE_VIEW (glade_xml_get_widget (xml, "custom_pager"));
    check_p (widget);

    gtk_tree_view_set_model (widget, cpse_list_new ());

    tree_view_insert_column (widget, COL_TEXT      , _("Name"));
    tree_view_insert_column (widget, COL_WIDTH     , _("Width"));
    tree_view_insert_column (widget, COL_HEIGHT    , _("Height"));
    tree_view_insert_column (widget, COL_RESOLUTION, _("Resolution"));
    tree_view_insert_column (widget, COL_NAME      , _("ID"));
  }
  {
    GtkFileChooser *widget;

    widget = GTK_FILE_CHOOSER (glade_xml_get_widget (xml, "file_browser"));
    check_p (widget);

    gtk_file_chooser_set_action (widget, GTK_FILE_CHOOSER_ACTION_OPEN);
  }
  {				/* "steal" a printer icon ;-) */
    GtkButton *b1 = GTK_BUTTON (glade_xml_get_widget (xml, "apply"));
    GtkButton *b2 = GTK_BUTTON (gtk_button_new_from_stock (GTK_STOCK_PRINT));
    GtkWidget *im = NULL;

    check_p (b1);

#if HAVE_GTK_COMBO_BOX_GET_ACTIVE_TEXT	/* then we also have this */
    if (b2) im = gtk_button_get_image (b2);
    if (im) gtk_button_set_image (b1, im);
#endif
    gtk_button_set_use_underline (b1, TRUE);

    if (b2) gtk_widget_destroy (GTK_WIDGET (b2));
  }

  /* connect the signals in the interface */
  glade_xml_signal_autoconnect (xml);

  /* start the event loop */
  gtk_main ();

  return (EXIT_SUCCESS);
}
