/*  message.c -- display warnings and ask y/N questions
    Copyright (C) 2008  SEIKO EPSON CORPORATION

    This file is part of "Custom Paper Size Editor".

    "Custom Paper Size Editor" is free software.
    You can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License or at your option any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY;  without even the implied warranty of FITNESS
    FOR A PARTICULAR PURPOSE or MERCHANTABILITY.
    See the GNU General Public License for more details.

    You should have received a verbatim copy of the GNU General Public
    License along with this program; if not, write to:

        Free Software Foundation, Inc.
        59 Temple Place, Suite 330
        Boston, MA  02111-1307  USA
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>

#include "message.h"

static const cpse_dialog_message msg[] = {
  { MSG_TYPE_YES_NO, MSG_CLOBBER_SETTINGS,
    N_("This action will discard your current settings.\n"
       "Continue?") },
  { MSG_TYPE_OKAY, MSG_BAD_FILE,
    N_("The PPD file you selected can not be edited with this utility.\n"
       "Possible causes include missing support for custom paper sizes"
       " and insufficient privileges to save or apply your changes.") },
  { MSG_TYPE_OKAY, MSG_NAME_EXISTS,
    N_("The name for this entry is already used or contains a forbidden"
       " character (one of ':', '<', '>').\n"
       "Please use another name.") },
  { MSG_TYPE_OKAY, MSG_TOO_MANY_NAMES,
    N_("Your changes could not be added.\n"
       "The maximum number of entries has been reached.") },
  { MSG_TYPE_OKAY, MSG_MISSING_ITEMS,
    N_("Your changes could not be added or applied.\n"
       "One or more data fields are missing.") },
  { MSG_TYPE_OKAY, MSG_NO_SELECTION,
    N_("Your changes can not be applied.\n"
       "You need to select an custom paper size entry first.") },
  { MSG_TYPE_YES_NO, MSG_REALLY_DELETE,
    N_("Really delete the selected entry?") },
  { MSG_TYPE_YES_NO, MSG_REALLY_SAVE,
    N_("This will overwrite the currently selected file.\n"
       "Continue?") },
  { MSG_TYPE_YES_NO, MSG_REALLY_APPLY,
    N_("This will modify the currently selected printer's settings.\n"
       "Continue?") },
  { MSG_TYPE_YES_NO, MSG_REALLY_SAVE_AND_APPLY,
    N_("This will change the selected file and the settings of the"
       " selected printer.\n"
       "Continue?") },
  { MSG_TYPE_OKAY, MSG_SAVE_FAILURE,
    N_("Failure trying to save your changes.\n") },
  { MSG_TYPE_OKAY, MSG_APPLY_FAILURE,
    N_("Failure trying to apply your changes.\n"
       "Make sure you have sufficient privileges to modify your printer"
       " settings.") },
  { MSG_TYPE_OKAY, MSG_APPLY_SUCCESS,
    N_("Successfully applied your changes.") },
  { MSG_TYPE_YES_NO, MSG_REALLY_QUIT,
    N_("There are unsaved and/or unapplied changes.\n"
       "Continue?") },
  { MSG_TYPE_OKAY, MSG_NO_PPDFILE_SELECTED,
    N_("You have not selected a file.  Your changes can only be applied"
       " to a printer.") },
  { MSG_TYPE_OKAY, MSG_NO_PRINTER_SELECTED,
    N_("You have not selected a printer.  Your changes can only be saved"
       " to file.") },
  { MSG_TYPE_OKAY, MSG_NO_PRINTERS_FOUND,
    N_("No printers found.") },
  { MSG_TYPE_YES_NO, MSG_LOAD_PPD_FILE,
    N_("Read a different PPD file for this printer?") },
  { MSG_TYPE_YES_NO, MSG_REGISTER_ENTRY,
    N_("Register?") },
};


#ifdef CPSE_GTK

gint
run_dialog (gint type, gint msg_id)
{
  GtkWidget *dialog = NULL;
  gint reply;
  gchar *text;

  require (type == msg[msg_id].type);

  if (MSG_TYPE_OKAY == type)
    dialog = glade_xml_get_widget (xml, "okay_dialog");
  else if (MSG_TYPE_YES_NO == type)
    dialog = glade_xml_get_widget (xml, "yes_no_dialog");
  else
    log_msg ("Request for unsupported dialog type: %d", type);

  require (dialog);

  text = g_markup_escape_text (_(msg[msg_id].text), -1);
  gtk_message_dialog_set_markup (GTK_MESSAGE_DIALOG (dialog), text);
  g_free (text);

  reply = gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_hide (dialog);

  return (reply);
}

#else  /* !defined (CPSE_GTK) */

#include <ctype.h>
#include <locale.h>
#include <stdio.h>

gint
run_dialog (gint type, gint msg_id)
{
  gchar *line = NULL;
  size_t size = 0;

  gint reply;

  require (type == msg[msg_id].type);
  require (MSG_TYPE_OKAY == type || MSG_TYPE_YES_NO == type);

  if (MSG_TYPE_OKAY == type)
    {
      reply = MSG_RESPONSE_OKAY;
      printf ("%s [Enter] ", _(msg[msg_id].text));
      getline (&line, &size, stdin);
    }
  else if (MSG_TYPE_YES_NO == type)
    {
      gboolean valid_input = FALSE;

      printf ("%s", _(msg[msg_id].text));
      do
	{
	  printf (" [y/N] ");
	  if (-1 != getline (&line, &size, stdin))
	    {
	      gchar *p = line;
	      gchar *l = setlocale (LC_CTYPE, "C");

	      while (p && *p && isspace (*p))
		++p;

	      if (p && 'Y' == toupper (*p))
		reply = MSG_RESPONSE_YES;

	      if (p && ('N' == toupper (*p)
			|| 0 == strlen (p)))
		reply = MSG_RESPONSE_NO;

	      setlocale (LC_CTYPE, l);

	      valid_input = (MSG_RESPONSE_YES == reply
			     || MSG_RESPONSE_NO == reply);
	    }
	}
      while (!valid_input);
    }
  else
    exit (EXIT_FAILURE);

  g_free (line);

  return (reply);
}

#endif /* !defined (CPSE_GTK) */
