/*  model.c -- 
    Copyright (C) 2008  SEIKO EPSON CORPORATION

    This file is part of "Custom Paper Size Editor".

    "Custom Paper Size Editor" is free software.
    You can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License or at your option any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY;  without even the implied warranty of FITNESS
    FOR A PARTICULAR PURPOSE or MERCHANTABILITY.
    See the GNU General Public License for more details.

    You should have received a verbatim copy of the GNU General Public
    License along with this program; if not, write to:

        Free Software Foundation, Inc.
        59 Temple Place, Suite 330
        Boston, MA  02111-1307  USA
 */


#include "model.h"


CPSEModel *
cpse_model_new (void)
{
  CPSEModel *m = NULL;

  m = g_malloc (sizeof (CPSEModel));
  if (!m) return (NULL);

  m->custom_paper = g_malloc (MAX_CUSTOM_PAPERS * sizeof (CPSEPaper));
  if (!m->custom_paper)
    {
      g_free (m);
      return (NULL);
    }

  /* members that may get checked by cpse_model_reset() */
  m->printer.model    = NULL;
  m->printer.ppd_file = NULL;
  m->dest_sz = 0;

  cpse_model_reset (m);
  return (m);
}

void
cpse_model_reset (CPSEModel *m)
{
  memset (m->custom_paper, 0, MAX_CUSTOM_PAPERS * sizeof (CPSEPaper));
  m->custom_paper_sz = MAX_CUSTOM_PAPERS;

  g_free ((gchar *) m->printer.model);
#ifndef CPSE_GTK
  g_free ((gchar *) m->printer.ppd_file);
#endif

  m->printer.model    = NULL;
  m->printer.name     = NULL;
  m->printer.ppd_file = NULL;
  m->printer.cups_ppd = NULL;

  m->units = UNIT_CM;

  m->has_hw_res   = FALSE;
  m->needs_update = FALSE;

  m->width.min = CPSE_MIN_WIDTH;
  m->width.max = m->width.min;
  m->width.val = m->width.max;

  m->height.min = CPSE_MIN_HEIGHT;
  m->height.max = m->height.min;
  m->height.val = m->height.max;

  m->resolution.min = CPSE_MIN_RESOLUTION;
  m->resolution.max = CPSE_MAX_RESOLUTION;
  m->resolution.val = CPSE_DEF_RESOLUTION;

  m->offset_width  = 0;
  m->offset_height = 0;

  if (0 != m->dest_sz)
    {
      cupsFreeDests (m->dest_sz, m->dest);
    }
  m->dest_sz = cupsGetDests (&(m->dest));
}

gdouble
cpse_model_convert (gdouble value, CPSEUnit from, CPSEUnit to)
{
  gdouble factor = -1;

  /* Ugly hacks to allow bypassing conversion for resolution values.
   */
  if (MAX_UNIT == from || MAX_UNIT == to)
    return value;

  require (0 <= from && from < MAX_UNIT);
  require (0 <= to   &&   to < MAX_UNIT);

  if (UNIT_CM == to)
    factor = CM_PER_INCH;
  else if (UNIT_INCH == to)
    factor = 1;
  else if (UNIT_MM   == to)
    factor = MM_PER_INCH;
  else if (UNIT_DOTS == to)
    factor = DOTS_PER_INCH;

  require (0 < factor);

  if (UNIT_CM == from)
    factor /= CM_PER_INCH;
  else if (UNIT_INCH == from)
    factor /= 1;
  else if (UNIT_MM   == from)
    factor /= MM_PER_INCH;
  else if (UNIT_DOTS == from)
    factor /= DOTS_PER_INCH;

  require (0 < factor);

  return (factor * value);
}

size_t
cpse_model_paper_count (const CPSEModel *m)
{
  size_t count = 0;

  require (m);

  while (count < m->custom_paper_sz
	 && m->custom_paper[count].text[0])
    ++count;

  return (count);
}

gboolean
cpse_model_is_duplicate (const CPSEModel *m, const gchar *text)
{
  gboolean found = FALSE;

  size_t count = 0;

  require (m);
  
  while (!found
	 && count < m->custom_paper_sz
	 && m->custom_paper[count].text[0])
    {
      found = (0 == strcmp (text, m->custom_paper[count].text));
      ++count;
    }

  return (found);
}


#ifdef CPSE_GTK

GtkTreeModel *
cpse_list_new (void)
{
  GtkListStore *list;

  list = gtk_list_store_new (MAX_COLUMNS, G_TYPE_STRING,
			     G_TYPE_DOUBLE, G_TYPE_DOUBLE, G_TYPE_DOUBLE,
			     G_TYPE_STRING);
  return (GTK_TREE_MODEL (list));
}

#endif /* CPSE_GTK */
