/*  ppd.c -- PPD file handling
    Copyright (C) 2008  SEIKO EPSON CORPORATION

    This file is part of "Custom Paper Size Editor".

    "Custom Paper Size Editor" is free software.
    You can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation;
    either version 2 of the License or at your option any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY;  without even the implied warranty of FITNESS
    FOR A PARTICULAR PURPOSE or MERCHANTABILITY.
    See the GNU General Public License for more details.

    You should have received a verbatim copy of the GNU General Public
    License along with this program; if not, write to:

        Free Software Foundation, Inc.
        59 Temple Place, Suite 330
        Boston, MA  02111-1307  USA
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <locale.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <zlib.h>

#include "ppd.h"

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

#if HAVE_GZDIRECT

#define ppd_direct	gzdirect

#else /* !HAVE_GZDIRECT */

/* Provide a backward compatibility hack for gzdirect.  The ppd_stream
   typedef only contains the minimal of members we need to implement
   our hack and should work with any of zlib 1.1.4, 1.2.1 and 1.2.2.
   As of 1.2.3, gzdirect should be provided by zlib itself.
 */
typedef struct
{
  z_stream stream;
  int      z_err;
  int      z_eof;
  FILE     *file;
  Byte     *inbuf;
  Byte     *outbuf;
  uLong    *crc;
  char     *msg;
  char     *path;
  int      transparent;
  char     mode;

} ppd_stream;

static int
ppd_direct (gzFile file)
{
  ppd_stream *s = (ppd_stream *) file;

  if (!s || s->mode != 'r') return (0);
  return (s->transparent);
}
#endif /* !HAVE_GZDIRECT */


#define PPD_RELAX_WHITESPACE_CHECKS	TRUE

enum {
  PPD_VALUE_NONE,
  PPD_VALUE_INVOCATION,
  PPD_VALUE_QUOTED,
  PPD_VALUE_SYMBOL,
  PPD_VALUE_STRING,
};

typedef struct
{
  const guchar *main_key;
  const guchar *opt_name;
  const guchar *opt_text;
  const guchar *value;
  const guchar *new_line;

  size_t main_key_len;
  size_t opt_name_len;
  size_t opt_text_len;
  size_t value_len;

  gint     value_type;
  gboolean continues;

} ppd_statement;

typedef struct
{
  const guchar *key;
  gboolean      seen;
  size_t        key_len;

} ppd_custom_page;

static ppd_custom_page key_word[] =
  {				/* other fields computed at run-time */
    { (const guchar *) "*PageSize"       },
    { (const guchar *) "*PaperDimension" },
    { (const guchar *) "*PageRegion"     },
    { (const guchar *) "*ImageableArea"  },
    { NULL }			/* terminator */
  };

static gboolean
ppd_error (ppd_file_t *ppd, const gchar *msg)
{
  if (msg) log_err ("%s", msg);
  if (ppd) ppdClose (ppd);
  return (FALSE);
}

static gboolean
ppd_get_attr_double (ppd_file_t *ppd, const gchar *name,
		     gdouble *value)
{
  ppd_attr_t *attr;

  if (!ppd || !name || !value)
    return (FALSE);

  attr = ppdFindAttr (ppd, name, NULL);
  if (!attr || !attr->value)
    return (FALSE);

  return (0 < sscanf (attr->value, "%lf", value));
}

/* Scans a bit of PostScript code for a pair of values in a format
   that is not necessarily anchored at the start of the code.  It
   returns the number of successfully scanned values matching the
   format.

   In case of error -1 is returned.  Note that EOF may be returned
   as well.
 */
static gint
ppd_scanf (const gchar *code, const gchar *format, gdouble *x, gdouble *y)
{
  const gchar *key_start;
  ptrdiff_t    key_length;

  if (!code || !format || !x || !y)
    return (-1);

  key_length = strchr (format, '%') - format;
  if (0 > key_length) return (-1);

  key_start = strchr (code, format[0]);
  while (key_start && 0 != strncmp (format, key_start, key_length))
    {
      key_start = strchr (++key_start, format[0]);
    }
  if (!key_start) return (-1);

  return (sscanf (key_start, format, x, y));
}

/*! Opens a PPD \a file for a \a printer.
 */
gboolean
ppd_open (const gchar *file, const gchar *printer, CPSEModel *model)
{
  gboolean result = TRUE;	/* be optimistic */

  ppd_file_t    *ppd;
  ppd_option_t  *option;
#if HAVE_PPDFINDCUSTOMPARAM
  ppd_coption_t *coption;
#endif

  CPSEPrinter trial;

  gdouble min_width = CPSE_MIN_WIDTH;
  gdouble max_width;		/* from PPD file */
  gdouble min_height = CPSE_MIN_HEIGHT;
  gdouble max_height;		/* from PPD file */
  gdouble min_resolution = CPSE_MIN_RESOLUTION;
  gdouble max_resolution = CPSE_MAX_RESOLUTION;

  gdouble offset_width  = 0;
  gdouble offset_height = 0;

  gboolean has_hw_res;		/* from PPD file */

  CPSEPaper *custom_paper;
  gint       custom_paper_sz;
  gint       custom_paper_count;

  gboolean file_change;
  gboolean printer_change;

  int i;
  ppd_choice_t *p;

  require ((file || printer) && model);

  trial.model    = NULL;
  trial.name     = printer;
  trial.ppd_file = file;
  trial.cups_ppd = NULL;

  if (trial.ppd_file && 0 == strlen (trial.ppd_file))
    {
      log_errno (ENOTSUP);
      return (FALSE);
    }

  {				/* figure out what has changed */
    file_change = (   (!model->printer.ppd_file &&  trial.ppd_file)
		   || ( model->printer.ppd_file && !trial.ppd_file)
		   || ( model->printer.ppd_file &&  trial.ppd_file
		       && 0 != strcmp (model->printer.ppd_file,
				       trial.ppd_file)));
    printer_change = (   (!model->printer.name &&  trial.name
			  && 0 != strlen (trial.name))
		      || ( model->printer.name && !trial.name
			  && 0 != strlen (model->printer.name))
		      || ( model->printer.name &&  trial.name
			  && 0 != strcmp (model->printer.name, trial.name)));
  }

  if (!(file_change || printer_change))
    return (TRUE);		/* no change, nothing to do */

  if (printer_change && !file_change)
    {
      if (trial.name && 0 < strlen (trial.name))
	{
	  trial.ppd_file = NULL;
	}
      else			/* unselecting printer */
	{
	  if (!model->printer.ppd_file)
	    {
	      int i = 0;

	      model->printer.model = NULL;

	      for (; i < MAX_CUSTOM_PAPERS; ++i)
		memset (model->custom_paper +i, '\0', sizeof (CPSEPaper));

	      if (model->printer.cups_ppd)
		unlink (model->printer.cups_ppd);
	    }
	  model->printer.name = NULL;
	  g_free (model->printer.cups_ppd);
	  model->printer.cups_ppd = NULL;
	  return (TRUE);
	}
    }
  else
    {
      trial.cups_ppd = NULL;
    }

  if (!trial.ppd_file)
    {
      trial.cups_ppd = g_strdup (cupsGetPPD (trial.name));
      if (!trial.cups_ppd)
	{
	  log_errno (ENOMEM);
	  return (FALSE);
	}
      ppd = ppdOpenFile (trial.cups_ppd);
    }
  else
    {
      ppd = ppdOpenFile (trial.ppd_file);
    }

  if (!ppd || !ppd->modelname)
    return (ppd_error (ppd, "Not a valid PPD file\n"));

  ppdMarkDefaults (ppd);

  /* Check whether we will require hardware resolution settings.
   */
  p = ppdFindMarkedChoice (ppd, "PageSize");
  if (!p || !p->code)
    return (ppd_error (ppd, "Missing required keyword\n"));

  has_hw_res = (NULL != strstr (p->code, "/HWResolution["));

  /* These are not required by the PPD spec but we require them for
     the use of a PPD file with CPSE.
   */
  if (   !ppd_get_attr_double (ppd, "MaxMediaWidth" , &max_width)
      || !ppd_get_attr_double (ppd, "MaxMediaHeight", &max_height))
    return (ppd_error (ppd, "Missing MaxMedia* fields.\n"));

#if HAVE_PPDFINDCUSTOMPARAM
  /* Grab custom page offsets if present in PPD file.
   */
  coption = ppdFindCustomOption (ppd, "PageSize");
  if (coption)
    {
      ppd_cparam_t *c;

      c = ppdFindCustomParam (coption, "WidthOffset");
      if (c) offset_width  = c->minimum.custom_points;
      c = ppdFindCustomParam (coption, "HeightOffset");
      if (c) offset_height = c->minimum.custom_points;
    }
#endif /* HAVE_PPDFINDCUSTOMPARAM */

  /* Grab all the PageSize definitions and setup memory for the user
     defined ones.
   */
  option = ppdFindOption (ppd, "PageSize");
  if (!option)
    return (ppd_error (ppd, "Missing PageSize options"));

  trial.model  = g_strdup (ppd->modelname);
  custom_paper = g_malloc0 (MAX_CUSTOM_PAPERS * sizeof (CPSEPaper));
  if (!trial.model || !custom_paper)
    return (ppd_error (ppd, "Out of memory"));

  custom_paper_sz = MAX_CUSTOM_PAPERS;

  /* Collect the user defined PageSize definitions.  We do full-blown
     validity checking and refuse to process PPD files with internal
     inconsistencies.
   */
  i = option->num_choices;
  p = option->choices;
  custom_paper_count = 0;
  for (; custom_paper_sz > custom_paper_count && 0 < i && p; --i, ++p)
    {
      gboolean ok = TRUE;	/* be optimistic, again */
      ppd_choice_t *choice;
      int cnt;
      gdouble w = 0;
      gdouble h = 0;
      gdouble x = 0;
      gdouble y = 0;

      choice = ppdFindChoice (option, p->choice);

      if (!(choice = ppdFindChoice (option, p->choice))
	  || (0 != strncmp (CUSTOM_PAPER_PREFIX, choice->choice,
			    strlen (CUSTOM_PAPER_PREFIX))))
	continue;

      cnt = ppd_scanf (choice->code, "/PageSize[%lf %lf]", &w, &h);
      ok = ok && (2 == cnt);
      ok = ok && (min_width  <= w && w <= max_width);
      ok = ok && (min_height <= h && h <= max_height);

      if (has_hw_res)
	{
	  cnt = ppd_scanf (choice->code, "/HWResolution[%lf %lf]", &x, &y);
	  ok = ok && (2 == cnt);
	  ok = ok && (min_resolution <= x && x <= max_resolution);
	  ok = ok && (min_resolution <= y && y <= max_resolution);
	  ok = ok && (x == y);
	}

      ok = ok && (MAX_CUSTOM_PAPERS > custom_paper_count);
      if (ok)
	{
	  int j = custom_paper_count;

	  strncpy (custom_paper[j].name, choice->choice, PPD_MAX_NAME);
	  strncpy (custom_paper[j].text, choice->text  , PPD_MAX_TEXT);
	  custom_paper[j].width = w;
	  custom_paper[j].height = h;
	  custom_paper[j].resolution = x;

	  ++custom_paper_count;
	}
      else
	{
	  log_msg ("Too many custom paper size definitions or PPD"
		   " assumption violated for '%s'", choice->text);
	}
      result = result && ok;
    }

  if (result)			/* finally, update the model */
    {
      g_free ((gchar *) model->printer.model);
      g_free (model->custom_paper);
#ifndef CPSE_GTK		/* avoid a CLI memory leak */
      g_free ((gchar *) model->printer.ppd_file);
#endif
      if (model->printer.cups_ppd)
	unlink (model->printer.cups_ppd);
      g_free (model->printer.cups_ppd);

      model->printer.model    = trial.model;
      model->printer.name     = trial.name;
      model->printer.ppd_file = trial.ppd_file;
      model->printer.cups_ppd = trial.cups_ppd;

      model->has_hw_res   = has_hw_res;
      model->needs_update = FALSE;

      model->width.min = min_width;
      model->width.max = max_width;
      model->width.val = max_width;

      model->height.min = min_height;
      model->height.max = max_height;
      model->height.val = max_height;

      model->resolution.min = min_resolution;
      model->resolution.max = max_resolution;
      model->resolution.val = CPSE_DEF_RESOLUTION;

      if (model->resolution.val > model->resolution.max)
	model->resolution.val = model->resolution.max;
      if (model->resolution.val < model->resolution.min)
	model->resolution.val = model->resolution.min;

      model->offset_width  = offset_width;
      model->offset_height = offset_height;

      model->custom_paper    = custom_paper;
      model->custom_paper_sz = custom_paper_sz;
    }
  else
    {
      g_free ((gchar *) trial.model);
      g_free (custom_paper);

      if (trial.cups_ppd)
	unlink (trial.cups_ppd);
      g_free (trial.cups_ppd);
    }

  ppdClose (ppd);

  return (result);
}

/*! Returns a file descriptor for a temporary file.

    The temporary file is opened with 0600 permissions.  The caller
    has to close the file.

    The name of the temporary file is returned via tmp_file.  The
    caller is responsible for releasing any allocated memory.

    If no temporary file could be opened, an invalid descriptor is
    return and the value of tmp_file will be set to \c NULL.  Any
    resources allocated so far will be released.
 */
static int
ppd_tmp_file (const gchar *ppd_file, gchar **tmp_file)
{
  int    fd;
  mode_t mask;

  if (!ppd_file || !tmp_file)
    {
      log_errno (EINVAL);
      return (-1);
    }
  if (*tmp_file)
    {
      log_dbg ("caller may cause memory leak");
    }

  {				/* create a pattern for the tmp_file */
    const gchar *xxxxxx = ".XXXXXX";
    const gchar *period = ".";

    ptrdiff_t dir_length;
    gint      tmp_length;

    dir_length = strrchr (ppd_file, G_DIR_SEPARATOR) + 1 - ppd_file;
    tmp_length = strlen (ppd_file) + strlen (period) + strlen (xxxxxx);
    ++tmp_length;			/* terminating NUL */
    *tmp_file = g_malloc (tmp_length * sizeof (gchar));
    if (!*tmp_file)
      {
	log_errno (errno);
	return (-1);
      }
    strncpy (*tmp_file, ppd_file, dir_length);	/* not NUL terminated! */
    strcpy (*tmp_file + dir_length, period);	/* now NUL terminated */
    strcat (*tmp_file, ppd_file + dir_length);
    strcat (*tmp_file, xxxxxx);

    if (tmp_length - 1 != strlen (*tmp_file))	/* sanity check */
      {
	log_err ("internal inconsistency");
	g_free (*tmp_file);
	return (-1);
      }
  }

  mask = umask (0077);		/* no prying eyes, please */
  fd = mkstemp (*tmp_file);
  if (0 > fd)
    {
      log_errno (errno);
      g_free (*tmp_file);
      *tmp_file = NULL;
    }
  umask (mask);

  return (fd);
}

/* FIXME: This is incorrect for '\r' using files.  The bug is sure to
   manifest itself when buf_sz is 1 on every invocation.  We really
   should be keeping some state across calls (or probably better yet,
   get state passed in for us to update) when we find '\r' at the end
   of buffer boundary.
 */
/* PPD spec: allows *any* combination of '\n' and '\r' as EOL!
   Implementation does not consider '\n\r'.
 */
/* BTW, normally an EOL ends a statement, but there are exceptions.
   These are noted on a case-by-case basis in the spec.
 */
/* FIXME: This is horribly broken for any multibyte encoding that can
   have '\n' and '\r' as there non-first byte.  I'm thinking UCS-2 and
   UCS-4.  AFAIU, all multi-byte non-vendor encodings should be okay
   though.
 */
static gboolean
find_eol (const guchar *buffer, size_t buf_sz, size_t *offset)
{
  gboolean eol = FALSE;
  size_t   cnt = 0;

  if (!buffer)
    return (FALSE);

  while (!eol && cnt < buf_sz)
    {
      if ('\n' == *buffer)
	eol = TRUE;
      else if ('\r' == *buffer)
	{
	  if (cnt + 1 < buf_sz)
	    {
	      if ('\n' == *(buffer+1)) ++cnt;
	      eol = TRUE;
	    }
	}
      ++buffer;
      ++cnt;
    }
  if (offset) *offset = cnt;

  return (eol);
}

/*! Tokenizes a syntactically valid line of a PPD file.

    Returns \c TRUE if the line is syntactically valid, \c FALSE
    otherwise.  Information passed back via the \a statement is only
    meaningful in when \c TRUE is returned.

    Note: we really need unsigned characters here as we will be doing
    numerical comparisons against decimal valued character codes.

    Note: Not sure whether main keyword handling is correct if a slash
    is encountered.  A slash is definitely forbidden as part of a main
    keyword, but it is not clear whether they are to be considered as
    valid main keyword terminators.  The implementation considers them
    as invalid.

    Note: The implementation uses the same length restriction as CUPS
    on translation strings for option keywords (opt_name).  There does
    not seem to be any mention of such a limit in the PPD spec though.

    Note: It is not clear which byte values are allowed in translation
    strings.  The implementation assumes that 8-bit bytes are allowed
    as it would otherwise be extremely cumbersome to do translations
    in a large number of languages.

    \bugs: Does not validate hexidecimal escape sequences.
 */
static gboolean
ppd_tokenize (const guchar *line, size_t line_sz, ppd_statement *statement)
{
  const guchar *p = line;
  size_t  i = 0;
  size_t  j = 0;

  if (!line) return (FALSE);
  if (PPD_MAX_LINE - 1 < line_sz) return (FALSE);
  if (!statement) return (FALSE);

  {				/* reset state */
    statement->main_key = NULL;
    statement->opt_name = NULL;
    statement->opt_text = NULL;
    statement->value    = NULL;
    statement->new_line = NULL;

    statement->main_key_len = 0;
    statement->opt_name_len = 0;
    statement->opt_text_len = 0;
    statement->value_len    = 0;
  }

  if (3 > line_sz)
    {				/* check for an empty line */
      if (1 == line_sz && ('\n' == line[0] || '\r' == line[0]))
	{
	  statement->new_line = line;
	  return (TRUE);
	}
      if (2 == line_sz && (   0 == memcmp ("\r\n", line, line_sz)
			   || 0 == memcmp ("\n\r", line, line_sz)))
	{
	  statement->new_line = line;
	  return (TRUE);
	}
    }
  if (2 > line_sz) return (FALSE);

  if (!statement->continues)
    {				/* check for a comment */
      if (0 == memcmp ("*%", line, strlen ("*%")))
	{
	  statement->new_line = line + line_sz - 2;
	  if ('\n' != *(statement->new_line) && '\r' != *(statement->new_line))
	    statement->new_line++;
	  return (TRUE);
	}
    }

  if (!statement->continues)
    {
      if (PPD_RELAX_WHITESPACE_CHECKS)	/* whitespace only lines */
	{
	  while (i < line_sz && ('\t' == *p || ' ' == *p))
	    ++i, ++p;
	  if (i < line_sz && ('\n' == *p || '\r' == *p))
	    {
	      statement->new_line = p;
	      return (TRUE);
	    }
	}
      if ('*' != *p) return (FALSE);

      statement->main_key = p;
      j = 0;
      while (j < PPD_MAX_NAME
	     && i < line_sz
	     && 32 < *p && *p < 127 && ':' != *p && '/' != *p)
	++j, ++i, ++p;
      if (PPD_MAX_NAME == j	/* keyword too long */
	  || 0 == j		/* no keyword! */
	  || line_sz == i	/* line too short */
	  || !memchr ("\t\n\r :", *p, strlen ("\t\n\r :")))
	{			/* or syntax error */
	  return (FALSE);
	}
      statement->main_key_len = j;
      if (memchr ("\t ", *p, strlen ("\t ")))
	{			/* need an option keyword */
	  while (i < line_sz && ('\t' == *p || ' ' == *p))
	    ++i, ++p;
	  if (PPD_RELAX_WHITESPACE_CHECKS)	/* lone main keyword */
	    {
	      if (i < line_sz && ('\n' == *p || '\r' == *p))
		{
		  statement->new_line = p;
		  return (TRUE);
		}
	    }
	  statement->opt_name = p;
	  j = 0;
	  while (j < PPD_MAX_NAME
		 && i < line_sz
		 && 32 < *p && *p < 127 && ':' != *p && '/' != *p)
	    ++j, ++i, ++p;
	  if (PPD_MAX_NAME == j	/* keyword too long */
	      || 0 == j		/* no keyword! */
	      || line_sz == i	/* line too short */
	      || !memchr (":/", *p, strlen (":/")))
	    {			/* or syntax error */
	      return (FALSE);
	    }
	  statement->opt_name_len = j;
	  if ('/' == *p)	/* need translation string, allow empty */
	    {
	      ++i, ++p;
	      statement->opt_text = p;
	      j = 0;
	      while (j < PPD_MAX_TEXT && i < line_sz
		     && 31 < *p && ':' != *p)
		++j, ++i, ++p;
	      if (PPD_MAX_TEXT == j
		  || line_sz == i
		  || ':' != *p)
		{
		  return (FALSE);
		}
	      statement->opt_text_len = j;
	    }
	}
      if (!memchr ("\n\r:", *p, strlen ("\n\r:")))	/* sanity check */
	{
	  log_err ("internal inconsistency");
	  return (FALSE);
	}
      if (':' == *p)		/* need a value */
	{
	  ++i, ++p;
	  while (i < line_sz && ('\t' == *p || ' ' == *p))
	    ++i, ++p;
	  if (line_sz == i || '\n' == *p || '\r' == *p)
	    return (FALSE);
	  statement->value = p;
	  j = 0;
	  switch (*p)
	    {
	    case '"':
	      statement->value_type
		= (statement->opt_name
		   && (0 != (memcmp ("*JCL", statement->opt_name,
				     strlen ("*JCL"))))
		   ? PPD_VALUE_INVOCATION
		   : PPD_VALUE_QUOTED);
	      break;
	    case '^':
	      statement->value_type = PPD_VALUE_SYMBOL;
	      break;
	    default:
	      statement->value_type = PPD_VALUE_STRING;
	    }
	  ++j, ++i, ++p;
	  switch (statement->value_type)
	    {
	    case PPD_VALUE_INVOCATION:
	      while (i < line_sz && '"' != *p
		     && ((31 < *p && *p < 127)
			 || ('\t' == *p || '\n' == *p || '\r' == *p)))
		++j, ++i, ++p;
	      if (line_sz == i)
		{
		  statement->new_line  = p;
		  statement->value_len = j;
		  statement->continues = TRUE;
		  return (TRUE);
		}
	      else if ('"' == *p)
		{
		  ++j, ++i, ++p;
		  statement->value_len = j;
		  statement->continues = FALSE;
		}
	      if ('/' == *p)	/* translation! */
		{
		  while (i < line_sz && (31 < *p || '\t' == *p))
		    ++j, ++i, ++p;
		  if ('\n' == *p || '\r' == *p)
		    {
		      statement->value_len = j;
		      statement->continues = FALSE;
		    }
		  else
		    return (FALSE);
		}
	      break;
	    case PPD_VALUE_QUOTED:
	      while (i < line_sz && '"' != *p
		     && (31 < *p || ('\t' == *p || '\n' == *p || '\r' == *p)))
		++j, ++i, ++p;
	      if (line_sz == i)
		{
		  statement->new_line  = p;
		  statement->value_len = j;
		  statement->continues = TRUE;
		  return (TRUE);
		}
	      else if ('"' == *p)
		{
		  ++j, ++i, ++p;
		  statement->value_len = j;
		  statement->continues = FALSE;
		}
	      if ('/' == *p)	/* translation! */
		{
		  while (i < line_sz && (31 < *p || '\t' == *p))
		    ++j, ++i, ++p;
		  if ('\n' == *p || '\r' == *p)
		    {
		      statement->value_len = j;
		      statement->continues = FALSE;
		    }
		  else
		    return (FALSE);
		}
	      break;
	    case PPD_VALUE_SYMBOL:
	      while (i < line_sz && 32 < *p && *p < 127)
		++j, ++i, ++p;
	      if ('\n' == *p || '\r' == *p)
		{
		  statement->value_len = j;
		  statement->continues = FALSE;
		}
	      break;
	    case PPD_VALUE_STRING:
	      while (i < line_sz
		     && ((31 < *p && *p < 127) || '\t' == *p)
		     && '/' != *p)
		++j, ++i, ++p;
	      if ('\n' == *p || '\r' == *p)
		{
		  statement->value_len = j;
		  statement->continues = FALSE;
		}
	      else if ('/' == *p)	/* translation string */
		{
		  while (i < line_sz && (31 < *p || '\t' == *p))
		    ++j, ++i, ++p;
		  if ('\n' == *p || '\r' == *p)
		    {
		      statement->value_len = j;
		      statement->continues = FALSE;
		    }
		  else
		    return (FALSE);
		}
	      break;
	    default:
	      log_err ("internal inconsistency");
	      return (FALSE);
	    }
	}
    }
  else
    {
      switch (statement->value_type)
	{
	case PPD_VALUE_INVOCATION:
	  statement->value = p;
	  while (i < line_sz &&  '"' != *p
		 && ((31 < *p && *p < 127)
		     || ('\t' == *p || '\n' == *p || '\r' == *p)))
	    ++j, ++i, ++p;
	  if (line_sz == i)
	    {
	      statement->new_line  = p;
	      statement->value_len = j;
	      statement->continues = TRUE;
	      return (TRUE);
	    }
	  else if ('"' == *p)
	    {
	      ++j, ++i, ++p;
	      statement->value_len = j;
	      statement->continues = FALSE;
	    }
	  break;
	case PPD_VALUE_QUOTED:
	  statement->value = p;
	  while (i < line_sz && '"' != *p
		 && (31 < *p || ('\t' == *p || '\n' == *p || '\r' == *p)))
	    ++j, ++i, ++p;
	  if (line_sz == i)
	    {
	      statement->new_line  = p;
	      statement->value_len = j;
	      statement->continues = TRUE;
	      return (TRUE);
	    }
	  else if ('"' == *p)
	    {
	      ++j, ++i, ++p;
	      statement->value_len = j;
	      statement->continues = FALSE;
	    }
	  break;
	default:
	  log_err ("internal inconsistency");
	  return (FALSE);
	}
    }

  if (PPD_RELAX_WHITESPACE_CHECKS)	/* trailing whitespace */
    {
      while (i < line_sz
	     && ('\t' == *p || ' ' == *p))
	++i, ++p;
    }

  if (line_sz == i) return (FALSE);

  if ('\n' == *p || '\r' == *p)
    {				/* check for valid EOL */
      statement->new_line = p;
      ++i, ++p;
      return (line_sz == i
	      || (line_sz == i + 1 && ('\n' == *p || '\r' == *p)));
    }

  return (FALSE);
}

/*! Tries hard to write all buf_sz bytes of a buffer.
 */
static size_t
fpwrite_loop (FILE *fp, const void *buffer, size_t buf_sz)
{
  int      fd = fileno (fp);

  while (0 < buf_sz)
    {
      ssize_t sz = write (fd, buffer, buf_sz);
      if (0 < sz)
	{
	  buffer += sz;
	  buf_sz -= sz;
	}
      else
	{
	  log_errno (errno);
	  return (buf_sz);
	}
    }
  return (buf_sz);
}

/*! Tries hard to write all buf_sz bytes of a buffer.
 */
static size_t
gzwrite_loop (gzFile gz, const void *buffer, size_t buf_sz)
{
  while (0 < buf_sz)
    {
      int sz = gzwrite (gz, buffer, buf_sz);
      if (0 < sz)
	{
	  buffer += sz;
	  buf_sz -= sz;
	}
      else
	{
	  log_errno (errno);	/* hoping this gets set */
	  return (buf_sz);
	}
    }
  return (buf_sz);
}

static size_t
write_line (const void *buffer, size_t buf_sz, FILE *fp, gzFile gz)
{
  if (fp)
    buf_sz = fpwrite_loop (fp, buffer, buf_sz);
  if (gz)
    buf_sz = gzwrite_loop (gz, buffer, buf_sz);

  return (buf_sz);
}

/*! Returns a pointer to the \c key_word entry matching statement \a s.

    Callers should check both the returned pointer as well as its \c
    key field to decide whether there was a match.  In case no match
    was found, the \c key will be \c NULL.

    The caller may modify fields of the entry returned.
 */
static ppd_custom_page *
get_custom_key (const ppd_statement *s)
{
  ppd_custom_page *p = key_word;

  gboolean found = FALSE;

  if (!p || !s || !s->main_key)
    return (NULL);

  do
    {
      found = (p->key_len == s->main_key_len
	       && 0 == memcmp (p->key, s->main_key, p->key_len));
    }
  while (!found && ++p && p->key);

  return (found ? p : NULL);
}

/*! Tells whether a statement \a s corresponds to a custom paper size.
 */
static gboolean
is_custom_page (const ppd_statement *s)
{
  return (s && s->opt_name
	  && 0 == memcmp (CUSTOM_PAPER_PREFIX, s->opt_name,
			  MIN (strlen (CUSTOM_PAPER_PREFIX),
			       s->opt_name_len)));
}

#define gustrcmp(s1,s2)	strcmp ((const gchar *) s1, (const gchar *) s2)

/*! Outputs a block of info for custom paper sizes.

    \bugs  Does not "word wrap" lines longer than 256 bytes as required
	   by the PPD spec.  It does output an error in case a line is
	   too long.  The lines _we_ create should be okay in virtually
	   all situations.
 */
static gboolean
splice_custom_pages (const guchar *key, const CPSEModel *model,
		     const guchar *new_line, FILE *fp, gzFile gz)
{
  gboolean ok = TRUE;		/* be optimistic */

  char *loc = NULL;

  gchar  buffer[PPD_MAX_LINE];
  size_t cutoff;
  size_t i;

  const gchar *nl = (const gchar *) new_line;

  const gchar *fmt;

  if (!key) return (FALSE);

  if (0 == gustrcmp (key, key_word[1].key))
    fmt = "%s %s/%s: \"%f %f\"";
  else if (0 == gustrcmp (key, key_word[3].key))
    fmt = "%s %s/%s: \"%f %f %f %f\"";
  else
    fmt = (model->has_hw_res
	   ? ("%s %s/%s: \"<</PageSize[%f %f]/HWResolution[%f %f]"
	      "/ImagingBBox null>>setpagedevice\"")
	   : ("%s %s/%s: \"<</PageSize[%f %f]"
	      "/ImagingBBox null>>setpagedevice\""));

  loc = setlocale (LC_NUMERIC, "C");
  for (i = 0; ok && i < MAX_CUSTOM_PAPERS; ++i)
    {
      const CPSEPaper *p = model->custom_paper + i;

      if (0 == strlen (p->name)) continue;

      if (0 == gustrcmp (key, key_word[1].key))
	cutoff = snprintf (buffer, PPD_MAX_LINE, fmt, key, p->name, p->text,
			   p->width, p->height);
      else if (0 == gustrcmp (key, key_word[3].key))
	cutoff = snprintf (buffer, PPD_MAX_LINE, fmt, key, p->name, p->text,
			   model->offset_width, model->offset_height,
			   p->width  - model->offset_width,
			   p->height - model->offset_height);
      else
	cutoff = (model->has_hw_res
		  ? snprintf (buffer, PPD_MAX_LINE, fmt, key, p->name, p->text,
			      p->width, p->height, p->resolution, p->resolution)
		  : snprintf (buffer, PPD_MAX_LINE, fmt, key, p->name, p->text,
			      p->width, p->height));
      ok = (PPD_MAX_LINE > cutoff + strlen (nl));
      if (ok)
	{
	  ok = (0 == write_line (buffer, cutoff, fp, gz)
		&& 0 == write_line (new_line, strlen (nl), fp, gz));
	}
      else
	{
	  log_errno (EOVERFLOW);
	}
    }
  setlocale (LC_NUMERIC, loc);

  return (ok);
}

/*! Updates a PPD file with the information from a model.

    We try to save to a file in the same directory with a '.'
    prefixed, and maybe some stuff suffixed to make sure it's unique.
    The CUPS code skips files with a leading period, see load_ppds()
    in scheduler/cups-driverd.c (scheduler/ppds.c for 1.1.x).

    The CUPS file API is not available until 1.2.0 and does not allow
    you to write compressed data even though it \e is implemented.
    So, we will be using libz for compressed output and all input.
    For uncompressed output we turn to write(), although we create a
    convenience stream with fdopen() just to distinguish the type of
    output we are using further down the line.

    If the file is successfully saved, an attempt will be made to set
    permissions and ownerships to that of the original.

    \bugs  Does not do error checking in error handling code.
 */
gboolean
ppd_save (const gchar *ppd_file, const CPSEModel *model)
{
  gboolean ok = TRUE;		/* be optimistic */

  gchar *org_file = NULL;
  gchar *tmp_file = NULL;
  int    tmp_file_d;

  gzFile i    = NULL;		/* input stream */
  gzFile o_gz = NULL;		/* compressed output stream */
  FILE  *o_fp = NULL;		/* uncompressed output stream */

  guchar gz_buf[BUFSIZ];	/* at least 256, usually bigger */
  int    gz_cnt;

  guchar buffer[PPD_MAX_LINE];
  size_t buf_sz;		/* how much of buffer is filled */

  ppd_statement statement;
  gboolean is_continuation_line;
  gboolean skipping_ppd_entry;

  if (!ppd_file || !model)
    {
      log_errno (EINVAL);
      return (FALSE);
    }

  /* Make sure we use the real original file because rename() does not
     honour symlinks.  It will happily replace a symlink with a file.
     We want the symlink to remain pointing to the updated file.
     Note that we rely on a glibc extension here, see realpath(3).
   */
  org_file = realpath (ppd_file, NULL);
  if (!org_file)
    {
      log_errno (errno);
      return (FALSE);
    }

  tmp_file_d = ppd_tmp_file (org_file, &tmp_file);
  if (0 > tmp_file_d)
    {
      g_free (tmp_file);
      g_free (org_file);
      return (FALSE);
    }

  i = gzopen (org_file, "r");
  if (!i)
    {
      close (tmp_file_d);
      unlink (tmp_file);
      g_free (tmp_file);
      g_free (org_file);
      return (FALSE);
    }

  if (ppd_direct (i))
    o_fp = fdopen (tmp_file_d, "w");
  else
    o_gz = gzdopen (tmp_file_d, "w");

  if (!(o_fp || o_gz))
    ok = FALSE;

  buf_sz = 0;
  memset (&statement, '\0', sizeof (statement));
  statement.continues = FALSE;
  skipping_ppd_entry  = FALSE;

  {				/* reset flags, init cache data */
    ppd_custom_page *p = key_word;

    while (p && p->key)
      {
	p->seen    = FALSE;
	p->key_len = strlen ((const gchar *const) p->key);
	++p;
      }
  }

  while (ok && (gz_cnt = gzread (i, gz_buf, BUFSIZ)))
    {
      guchar *gz_ptr = gz_buf;

      while (ok && 0 < gz_cnt)
	{
	  gboolean eol = FALSE;
	  size_t   max = PPD_MAX_LINE - 1 - buf_sz;
	  size_t   cnt = 0;

	  if (gz_cnt < max) max = gz_cnt;

	  eol = find_eol (gz_ptr, max, &cnt);

	  if (!eol && gz_cnt > max)
	    {
	      log_err ("file violates PPD spec!");
	      ok = FALSE;
	      break;
	    }

	  memcpy (buffer + buf_sz, gz_ptr, cnt);
	  gz_ptr += cnt;
	  gz_cnt -= cnt;
	  buf_sz += cnt;
	  buffer[buf_sz] = '\0';

	  is_continuation_line = statement.continues;

	  if (eol && (ok = ppd_tokenize (buffer, buf_sz, &statement)))
	    {
	      if (!is_continuation_line)
		{
		  ppd_custom_page *p = get_custom_key (&statement);

		  if (p && p->key)
		    {
		      if (!p->seen)
			{
			  ok = splice_custom_pages (p->key, model,
						    statement.new_line,
						    o_fp, o_gz);
			  p->seen = TRUE;
			}
		      skipping_ppd_entry = is_custom_page (&statement);
		    }
		}
	      if (!skipping_ppd_entry)
		{
		  ok = ok && (0 == write_line (buffer, buf_sz, o_fp, o_gz));
		}
	      else
		{
		  skipping_ppd_entry = statement.continues;
		}
	      buf_sz = 0;
	    }
	}				/* end of PPD line loop */
    }					/* end of gzread() loop */
  if (ok && 0 < buf_sz && !skipping_ppd_entry)
    {				/* flush last line for '\r' files */
      ok = (0 == write_line (buffer, buf_sz, o_fp, o_gz));
      buf_sz = 0;
    }

  if (ok)
    {
      struct stat ppd_stat;

      /* Try to change permissions, owner and group to match that of
	 the original file.  Log failures but continue nevertheless.
       */
      if (0 != stat (org_file, &ppd_stat))
	log_errno (errno);
      else
	{
	  if (0 != chmod (tmp_file, ppd_stat.st_mode))
	    log_errno (errno);
	  if (0 != chown (tmp_file, ppd_stat.st_uid, ppd_stat.st_gid))
	    log_errno (errno);
	}
      rename (tmp_file, org_file);
    }

  if (o_gz) gzclose (o_gz);
  if (o_fp) fclose (o_fp);
  gzclose (i);				/* closes tmp_file_d too */
  if (!ok) unlink (tmp_file);
  g_free (tmp_file);
  g_free (org_file);

  return (ok);
}

/*! Replaces the PPD file for a printer with the one passed.

    NOTE: maybe this should be reimplemented to talk to the CGI that's
    listening on http://localhost:631/admin.
 */
gboolean
ppd_apply (const gchar *ppd_file, const gchar *printer)
{
  const gchar *sh_cmd_fmt = "lpadmin -p %s -P %s";

  gchar *sh_cmd;
  gint   sh_cmd_length;
  gint   result;

  if (!ppd_file || !printer)
    {
      log_errno (EINVAL);
      return (FALSE);
    }

  sh_cmd_length = strlen (sh_cmd_fmt) + strlen (ppd_file) + strlen (printer);
  ++sh_cmd_length;		/* terminating NUL */
  sh_cmd = g_malloc (sh_cmd_length * sizeof (gchar));
  if (!sh_cmd)
    {
      log_errno (ENOMEM);
      return (FALSE);
    }

  if (sh_cmd_length <= snprintf (sh_cmd, sh_cmd_length,
				 sh_cmd_fmt, printer, ppd_file))
    {
      log_errno (ERANGE);
      return (FALSE);
    }

  if (getenv ("CPSE_DEVEL"))	/* just show what would be done */
    {
      log_dbg ("system (\"%s\");", sh_cmd);
      g_free (sh_cmd);
      return (FALSE);
    }

  result = system (sh_cmd);
  if (-1 == result)
    log_errno (errno);
  g_free (sh_cmd);

  return (!WEXITSTATUS (result));
}

/* We don't support hex-escaping so we can't have any of ":<>" in the
   user provided option texts.
 */
gboolean
ppd_is_valid_text (const gchar *text)
{
  return (text
	  && 0 < strlen (text)
	  && PPD_MAX_TEXT > strlen (text)
	  && strlen (text) == strcspn (text, ":<>"));
}
