/*! \file common.c

  \brief rastertoeplq common utility.

  \author
    Copyright (C) Seiko Epson Corporation 2007.

    Driver for EPSON LQ Series SIDM Printers

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdarg.h>
#include <ctype.h>
#include <time.h>

#include "common.h"

/*! \brief (for DEBUG only)

  Output delay count in micro sec.
  \note Set 50000 (50ms delay), When job cancel test.
 */
#define  OUTPUT_DELAY 0

/*! \brief (for DEBUG only)

  Mamory allocate counter */
static int alloc_counter = 0;

/*! \brief (for DEBUG only)

  Mamory free counter */
static int free_counter = 0;

/* ****************************************************************
   standard C function wrapper
   **************************************************************** */

/*! \brief formated text message output to printer.

    Like a printf(3).

    \param  fmt    Formated message string, like a printf(3).
    \param  ...    Argument params for \a fmt, like a printf(3).
    \return        Output status.
    \retval 0      completed
    \retval -1     write error
 */
FUNCTION int printf_out( const char *fmt, ... )
{
  va_list ap;
  int n = 0;
  int i = 0;
  char buf[MAX_STR_BUF] = {0,};

  ASSERT(NULL != fmt);  if ( NULL == fmt ) { return 0; }

  va_start(ap, fmt);
  if ( 0 != MYDEBUG ) {
    n = vsnprintf(buf, MAX_STR_BUF, fmt, ap);
  } else {
    n = vfprintf(stdout, fmt, ap);
  }
  va_end(ap);

  if ( 0 != MYDEBUG ) {
    for ( i = 0; i < n; i++ ) {
      if ( ( 0 == i ) && ( ESC == buf[i] ) ) {
	dbgprintf("\n ESC ");
      } else
	if ( isprint(buf[i]) ) {
	  dbgprintf("%c", buf[i]);
	} else {
	  dbgprintf(" %02x", buf[i]);
	}
    }
  }

  if ( 0 > n ) {
    return -1;
  }

  return 0;
}

/*! \brief binary data output.

    Binary data output to printer.
    like a fwrite(3).

    \param  data binary data.
    \param  size output binary data size in Bytes.
    \return 0    completed
    \retval -1   write error.
 */
FUNCTION int printd_out( unsigned char *data, size_t size )
{
  size_t writed = 0;
  size_t n = 0;
  int i = 0;

  ASSERT(NULL != data); if ( NULL == data ) { return 0; }
  ASSERT(size);         if ( !size )        { return 0; }

  if ( 0 != MYDEBUG ) {
    for ( i = 0; i < size; i++ ) {
      dbgprintf(" %02x", data[i]);
      if ( (2 > MYDEBUG) && (16 <= i) ) {
	dbgprintf("...");
	break;
      }

      if ( OUTPUT_DELAY ) {	/* (for job cancel test = 50ms) */
	if ( safe_usleep(OUTPUT_DELAY) ) {
	  dbgprintf("\n******** safe_usleep error ********\n");
	  return -1;
	}
      }

    }
    dbgprintf("\n");
    n = i;
  } else {
    n = 0;
    for ( i = 0; (3 > i) && (n < size); i++ ) {	/* retry limit = 3 */
      writed = fwrite(&data[n], sizeof(unsigned char), (size - n), stdout);
      n += writed;

      if ( OUTPUT_DELAY ) {	/* (for job cancel test = 50ms) */
	if ( safe_usleep(OUTPUT_DELAY) ) {
	  dbgprintf("\n******** safe_usleep error ********\n");
	  return -1;
	}
      }

    }
    if ( n != size ) {
      return -1;
    }
  }
  return 0;
}

/*! \brief Check all zero value in the memory.
  \param buff     one line buffer.
  \param num      number of bytes in \a buff
  \return         empty or not.
  \retval 1       empty
  \retval 0       not empty
 */
FUNCTION int is_empty( const unsigned char *buff, unsigned int num )
{
  unsigned int i;

  ASSERT(buff);

  for ( i = 0; i < num; i++ ) {
    if ( 0 != buff[i] ) {
      return 0;
    }
  }
  return 1;
}

/*! \brief Local round function.
  \param value  original value in \a from base.
  \param from   base unit for \a value.
  \param to     base unit for return value.
  \return       converted value from \a to base unit.
 */
FUNCTION unsigned long my_round_ul( unsigned long value, int from, int to )
{
  unsigned long retval = 0;

  ASSERT(from);
  ASSERT(to);

  if ( !value || !from || !to ) {
    return (unsigned long)0;
  }

  retval = value * to / from;
  if ( ( (value * 10 * to / from) % 10 )>= 5 ) {
    retval = retval + 1;
  }
  return retval;
}



/*! \brief malloc(3) wrapper function.

    see man malloc(3)
 */
FUNCTION void *my_malloc( size_t size )
{
  void *ptr = NULL;

  alloc_counter++;

  ASSERT(size); if ( !size ) { return NULL; }

  ptr = malloc(size);

  return ptr;
}

/*! \brief free(3) wrapper function.

    see man free(3)
 */
FUNCTION void my_free( void *ptr )
{
  free_counter++;

  ASSERT(NULL != ptr); if ( NULL == ptr ) { return; }

  free(ptr);
}


/*! \brief Output message to CUPS log file.

  \param file	(for DEBUG only) Source filename
  \param line	(for DEBUG only) Line number in the source file
  \param func   (for DEBUG only) Function name at the source code
  \param level  Log level for CUPS.
  \param fmt    Formated message string, like a printf(3).
  \param ...    Argument params for \a fmt, like a printf(3).
 */
FUNCTION void log_msg( const char *file, const int line, const char *func, const char *level, const char *fmt, ... )
{
  va_list ap;

  ASSERT(NULL != fmt);  if ( NULL == fmt ) { return; }

  fprintf(stderr, "%s", level);

  if(MYDEBUG){
    fprintf(stderr, "%s: %d: %s: ", file, line, func);
  }

  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
}

/*! \brief (for DEBUG only) */
FUNCTION void debug_check_memory_allocate_counter( void )
{
  if ( alloc_counter != free_counter ) {
    dbgprintf("!error: alloc_counter != free_counter\n");
    dbgprintf("    alloc_counter = %d\n", alloc_counter);
    dbgprintf("     free_counter = %d\n",  free_counter);
  }
}

/*! Micro seconds sleep
  \param  usec  wait count (micro second)
  \return       error status
  \retval 0     completed
  \retval other error
 */
FUNCTION int safe_usleep( const unsigned long usec )
{
  int retval;
  struct timeval tv;

  if ( 0 == usec )
    return 0;

  tv.tv_sec = usec / 1000000;
  tv.tv_usec = usec % 1000000;

  retval = select(0, NULL, NULL, NULL, &tv);

  return(retval);
}

/* end of file */
