/*! \file common.h

  \brief rastertoeplq common utility.

  \author
    Copyright (C) Seiko Epson Corporation 2007.

    Driver for EPSON LQ Series SIDM Printers

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#ifndef   common_h
#define   common_h

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <stdlib.h>
#include <cups/cups.h>
#include <cups/ppd.h>
#include <cups/raster.h>


/*! \brief (for DEBUG only)

  debug switch

  0 debug off\n
  1 debug on
*/
#define MYDEBUG     0


#define FUNCTION		/*!< Global function mark. */
#define LOCALFUNC static	/*!< File scope local function mark. */

#define LF  0x0a		/*!< Line feed command. */
#define FF  0x0c		/*!< Form feed command. */
#define CR  0x0d		/*!< Caridge return command. */
#define ESC 0x1b		/*!< Escape command for ESC/P. */
#define FS  0x1c		/*!< Escape command for ESC/P. */

#define MAX_STR_BUF 1024	/*!< Popular string buffer size in Byte. */

/*! \brief option list */
typedef struct option_choice_s {
  const char *key;		/*!< Option keyword, like a "Quality" */
  const ppd_choice_t *choice;	/*!< CUPS ppd_choice_t for \a key */
} option_choice_t;

/*! \brief Job Information parameters */
typedef struct job_info_s {
  /* job informations */
  option_choice_t  *options;	/*!< Current option settings */
  ppd_file_t       *ppd;	/*!< PPD file */

  /* page infomations */
  int               page_num;	/*!< Current page number */
  const ppd_size_t *page_size;	/*!< CUPS page size parameters */
  cups_raster_t    *ras;	/*!< Raster stream for printing */
  unsigned char    *input_raster_buff; /*!< Input CUPS raster data buffer (one line) */
  unsigned char    *band_buf;	/*!< Input CUPS raster data buffer (one band) */
  unsigned char    *bitmap_band_buf; /*!< Output bitmap data buffer (one band) */
  int               top_shift;	/*!< Top image shift (in line) */
  int               left_shift;	/*!< Left image shift (in pixel) */

  /* band image infomation */
  unsigned int      bmp_data_size; /*!< ESC * command bitmap data size (for job cancel) */
} job_info_t;


FUNCTION int printf_out( const char *fmt, ... );
FUNCTION int printd_out( unsigned char *data, size_t size );
FUNCTION int is_empty( const unsigned char *buff, unsigned int num );
FUNCTION unsigned long my_round_ul( unsigned long value, int from, int to );

FUNCTION void *my_malloc( size_t size );
FUNCTION void my_free( void *ptr );
FUNCTION void debug_check_memory_allocate_counter( void );
FUNCTION void log_msg( const char *file, const int line, const char *func, const char *level, const char *fmt, ... );
FUNCTION int safe_usleep( const unsigned long usec );


#define dbg __FILE__,__LINE__,__FUNCTION__
/*! \def dbg
  Debug information.

  \see \a log_msg
*/

#define LOGMSG    dbg,""	  	/*!< for CUPS log */
#define LOGPAGE   dbg,"PAGE: "	  	/*!< for CUPS log */
#define LOGERROR  dbg,"ERROR: "	  	/*!< for CUPS log */
#define LOGWARN   dbg,"WARNING: " 	/*!< for CUPS log */
#define LOGINFO   dbg,"INFO: "	  	/*!< for CUPS log */
#define LOGDEBUG1 dbg,"DEBUG: "	  	/*!< for CUPS log */
#define LOGDEBUG2 dbg,"DEBUG2: "  	/*!< for CUPS log */


/*! \def ASSERT
  (for DEBUG only)
 */
#if        (0 == MYDEBUG)
#define ASSERT(a) do{}while(0)
#else   /* (0 != MYDEBUG) */
#define ASSERT(a) do{ if(!(a)){ log_msg(LOGERROR,"***** ASSERT *****\n"); if(MYDEBUG) exit(1);} }while(0)
#endif  /* (0 != MYDEBUG) */

/*! \def dbgprintf
  (for DEBUG only)
 */
#if        (0 == MYDEBUG)
#define dbgprintf(...) do{}while(0)
#else   /* (0 != MYDEBUG) */
#define dbgprintf printf
#endif  /* (0 != MYDEBUG) */


#endif /* common_h */
/* end of file */
