/*! \file rastertoeplq.c

  \brief ESC/P2 filter for CUPS raster data.

  \author
    Copyright (C) Seiko Epson Corporation 2007,2008.2010

    Driver for EPSON LQ Series SIDM Printers

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <signal.h>

#include "common.h"
#include "rastertoeplq.h"

/*
 * Define...
 */

/*! \brief Base design unit in this driver */
#define DESIGNBASE  3600	/*!< Design base unit = 1/3600 dpi  */

/*! \brief Horizontal move base unit */
#define HMOVEBASE    360	/*!< unit = 1/360 dpi */

/*! \brief Virtical move base unit */
#define VMOVEBASE    360	/*!< unit = 1/360 dpi */


/*! \brief CUPS printer model number. */
#define EPSON_24PIN     1	/*!< EPSON 24 pins serial impact dot matrics printer. */
/*
#define EPSON_9PIN   0            non-support
*/

/*! \brief Tot shift image offset MAX value */
#define TSHIFTMAX     25.4	/*!< unit = mm */

/*! \brief Tot shift image offset MIN value */
#define TSHIFTMIN    -25.4	/*!< unit = mm */

/*! \brief Left shift image offset MAX value */
#define LSHIFTMAX     25.4	/*!< unit = mm */

/*! \brief Left shift image offset MIN value */
#define LSHIFTMIN    -25.4	/*!< unit = mm */


/* PPD Keywords */
#define PPDKEYPageSize    "PageSize"	/*!< PPD keyword for Medi Size */
#define PPDKEYResolution  "Resolution"	/*!< PPD keyword for Device Resolution */
#define PPDKEYQuality     "EPQuality"	/*!< EPSON private PPD keyword for Print Quality */
#define PPDKEYTopMP       "EPTopMP"	/*!< EPSON private PPD keyword for Top Margin Priority */
#define PPDKEYVShift      "EPVShift"	/*!< EPSON private PPD keyword for Top image Shift  */
#define PPDKEYHShift      "EPHShift"	/*!< EPSON private PPD keyword for Left image Shift  */
#define PPDKEYSpeed       "EPSpeed"		/*!< EPSON private PPD keyword for Print Speed */
#define PPDKEYPacketMode  "EPPackMode"	/*!< EPSON private PPD keyword for Packet Mode */

/*! \brief Image shift function disable */
#define NOSUPPOT_IMAGESHIFT 	/*!< disable image shift function. Please delete this define, when you want the function. */

/*
 * Prototypes...
 */

LOCALFUNC int start_job( job_info_t *job_info, const ppd_file_t *ppd, const int num_options, const cups_option_t *options );
LOCALFUNC int end_job( job_info_t *job_info );
LOCALFUNC int start_page( job_info_t *job_info, const ppd_file_t *ppd, const cups_page_header_t *page_header );
LOCALFUNC int end_page( job_info_t *job_info );
LOCALFUNC void cancel_job(int sig);
LOCALFUNC int line_feed( const cups_page_header_t *page_header, unsigned short *skip_line_num );
LOCALFUNC int send_band( job_info_t *job_info, const cups_page_header_t *page_header, unsigned short *skip_line_num, unsigned short *band_line, unsigned char *band_buf );
LOCALFUNC int band_out( job_info_t *job_info, unsigned char *bmpbuf, int mode, int band_bytes, int start, int length, int xreso );
LOCALFUNC int get_bitimage_mode( unsigned int xreso, unsigned int yreso );
LOCALFUNC ppd_file_t *cast_const_to_ppd_file_t ( const ppd_file_t *p );
LOCALFUNC ppd_option_t *cast_const_to_ppd_option_t( const ppd_option_t *p );
LOCALFUNC const char *find_option_value( const char *key, const int num_options, const cups_option_t *options );
LOCALFUNC const char *find_option_choice( const char *key, int num_options, const option_choice_t *options );
LOCALFUNC int is_match_key_value( const char *key, const char *value, int num_options, const option_choice_t *options );
LOCALFUNC int get_ppd_choice( option_choice_t *new_options, const ppd_file_t *ppd );
LOCALFUNC int overwrite_option_by_commandline_choice( option_choice_t *new_options, const ppd_file_t *ppd, const int num_options, const cups_option_t *options );


/*
 * File scope global parameters...
 */

/*! \brief option key list */
static const char option_key[][PPD_MAX_NAME] = {
  PPDKEYPageSize,
  PPDKEYResolution,
  PPDKEYQuality,
#ifndef NOSUPPOT_IMAGESHIFT
  PPDKEYVShift,
  PPDKEYHShift,
#endif  /* NOSUPPOT_IMAGESHIFT */
  PPDKEYTopMP,
  PPDKEYPacketMode,
  PPDKEYSpeed
};

/*! \brief number of items in option_key list. */
static const int option_key_num = sizeof(option_key) / PPD_MAX_NAME;

/*! \brief Job Information (real memory) */
job_info_t  _job_info = {0,};

/*! \brief Job Information pointer (access pointer) */
job_info_t  *job_info = &_job_info;

/*! \brief PrintSpedd parameter */
char printSpeed[MAX_STR_BUF] ={0};

/*
 * MAIN PROCESS...
 */


/*! \brief rastertoeplq

  main process.

  \param    fd            a input file descriptor.
  \param    num_options	  number of items in \a options.
  \param    options	  CUPS option list.
  \return                 error status
  \retval   zero          completed
  \retval   other         error
*/
FUNCTION  int rastertoeplq( int fd, int num_options, cups_option_t *options )
{
  cups_page_header_t       _page_header; /* Page header (real memory) */
  const cups_page_header_t *page_header = &_page_header; /* Page header (pointer) */
  unsigned short 	    skip_line_num = 0; /* number of skip lines */
  unsigned short            band_line = 0;  /* number of line in band_buf */
  int                       scan_line = 0;  /* Current line */
  int                       status = 0;	    /* temporary error status */
  int                       ret_status = 0; /* return status */
  int                       last_band = 0;  /* last band flag, 1=last band */
  int                       i = 0;

  dbgprintf("num_options = %d\n", num_options);
  for ( i = 0; i < num_options; i++ ) {
    dbgprintf("name=\"%s\"\tvalue=\"%s\"\n", options[i].name, options[i].value);
  }

  /* clear job_info */
  memset(job_info, 0, sizeof(job_info_t));

  job_info->ppd = ppdOpenFile(getenv("PPD"));

  if ( NULL == job_info->ppd ) {
    log_msg(LOGDEBUG1, "Can not read PPD information.\n");
    return ENOENT;
  }

  job_info->ras = cupsRasterOpen(fd, CUPS_RASTER_READ);
  ASSERT(job_info->ras);

  ret_status = start_job(job_info, job_info->ppd, num_options, options);
  if ( ret_status ) {
    log_msg(LOGDEBUG1, "Can not start job. [%d]\n", ret_status);
    cupsRasterClose(job_info->ras);
    job_info->ras = NULL;
    return ret_status;
  }

  /* page loop start */
  while (cupsRasterReadHeader(job_info->ras, &_page_header)) {

    /* count up page counter */
    job_info->page_num++;
    log_msg(LOGPAGE, "%d %d\n", job_info->page_num, page_header->NumCopies);

    if ( CUPS_CSPACE_K != page_header->cupsColorSpace ) {
      log_msg(LOGERROR, "CUPS data format error.\n");
      ret_status = -1;
      goto er_rastertoeplq1;
    }

    ret_status = start_page(job_info, job_info->ppd, page_header);

    if ( 0 != ret_status ) {
      goto er_rastertoeplq1;
    }


    last_band = 0;
    band_line = 0;
    if ( 0 < job_info->top_shift ) {
      skip_line_num = job_info->top_shift;
    } else {
      skip_line_num = 0;
    }


    /* loop start */
    for (scan_line = 0; scan_line < page_header->cupsHeight; scan_line++) {

      if ((scan_line & 127) == 0) {
	log_msg(LOGINFO, "Printing page %d, %d%% complete...\n", 
		job_info->page_num, 100 * scan_line / page_header->cupsHeight);
      }


      if ( 1 > cupsRasterReadPixels(job_info->ras, job_info->input_raster_buff, page_header->cupsBytesPerLine) ) {
	break;
      }

      /*
       * Judge last band...
       */
      if ( scan_line == (page_header->cupsHeight - job_info->top_shift - page_header->cupsRowCount) ) {

	last_band = 1;

	if ( 0 != band_line ) {
	  if ( 0 != skip_line_num ) {
	    ret_status = line_feed(page_header, &skip_line_num);
	    if ( 0 != ret_status ) {
	      goto er_rastertoeplq1;
	    }
	  }
	  ret_status = send_band(job_info, page_header, &skip_line_num, &band_line, job_info->band_buf);
	  if ( 0 != ret_status ) {
	    goto er_rastertoeplq1;
	  }
	}
      }

      if ( ( (0 > job_info->top_shift) && (scan_line < (job_info->top_shift * (-1)) ) ) ||
	   ( (0 < job_info->top_shift) && (scan_line >= (page_header->cupsHeight - job_info->top_shift) ) ) ) {
	;			/* do nothing */
      } else {

	if ( !last_band ) {

	  if ( ( 0 == band_line ) &&
	       is_empty(job_info->input_raster_buff, page_header->cupsBytesPerLine) ) {

	    skip_line_num++;

	  } else {

	    if ( 0 != skip_line_num ) {
	      ret_status = line_feed(page_header, &skip_line_num);
	      if ( 0 != ret_status ) {
		goto er_rastertoeplq1;
	      }
	    }

	    ASSERT( band_line < page_header->cupsRowCount );

	    memcpy(&(job_info->band_buf[band_line * page_header->cupsBytesPerLine]),
		   job_info->input_raster_buff, page_header->cupsBytesPerLine);
	    band_line++;

	    if ( page_header->cupsRowCount == band_line ) {

	      ret_status = send_band(job_info, page_header, &skip_line_num, &band_line, job_info->band_buf);
	      if ( 0 != ret_status ) {
		goto er_rastertoeplq1;
	      }

	    }
	  }
	} else {		/* last_band */

	  ASSERT( band_line < page_header->cupsRowCount );

	  memcpy(&(job_info->band_buf[band_line * page_header->cupsBytesPerLine]),
		 job_info->input_raster_buff, page_header->cupsBytesPerLine);
	  band_line++;

	} /* last_band */
      }

      dbgprintf("scan_line=%d", scan_line);
      dbgprintf(", last_band=%d", last_band);
      dbgprintf(", band_line=%d", band_line);
      dbgprintf(", skip_line_num=%d", skip_line_num);
      dbgprintf("\n");
    } /* loop end */

    ASSERT(band_line <= page_header->cupsRowCount);
    if ( band_line < page_header->cupsRowCount ) {
      memset(&job_info->band_buf[band_line * page_header->cupsBytesPerLine], 0,
	     page_header->cupsBytesPerLine * (page_header->cupsRowCount - band_line));
    }

    if ( !is_empty(job_info->input_raster_buff, page_header->cupsBytesPerLine * band_line ) ) {

      if ( 0 != skip_line_num ) {

	ret_status = line_feed(page_header, &skip_line_num);
	if ( 0 != ret_status ) {
	  goto er_rastertoeplq1;
	}
      }

      ret_status = send_band(job_info, page_header, &skip_line_num, &band_line, job_info->band_buf);
      if ( 0 != ret_status ) {
	goto er_rastertoeplq1;
      }
    }

    band_line = 0;
    skip_line_num = 0;

  er_rastertoeplq1:

    status = end_page(job_info);
    if ( !ret_status ) {
      ret_status = status;
    }

    if ( 0 != ret_status ) {
      break;
    }
  }

  status = end_job(job_info);
  if ( !ret_status ) {
    ret_status = status;
  }

  if (job_info->page_num == 0) {
    log_msg(LOGERROR, "No pages found!\n");
  } else {
    log_msg(LOGINFO, "Ready to print.\n", stderr);
  }

  debug_check_memory_allocate_counter();

  dbgprintf("\nall done.\n");

  return ret_status;
}

/*! \brief Start Job
  \param  job_info      job information
  \param  ppd           PPD information
  \param  num_options   number of items in \a options.
  \param  options       CUPS option list.
  \return               error status
  \retval 0             completed
  \retval other         error
 */
LOCALFUNC int start_job( job_info_t *job_info, const ppd_file_t *ppd, const int num_options, const cups_option_t *options )
{
  option_choice_t    *new_options = NULL;
  const ppd_size_t   *pageSize = NULL;
  int                 cups_model_number = -1; /* Model number */
  int                 status = 0;
  int                 i = 0;
  const char           *tmp_cp1 = NULL;
  int 					write_error = 0;

  ASSERT( NULL != ppd );
  if ( NULL == ppd ) {
    return EINVAL;
  }

  cups_model_number = ppd->model_number;

  if ( EPSON_24PIN != cups_model_number ) {
    log_msg(LOGERROR, "The printer model is not support.\n");
    return -1;
  }

  new_options = my_malloc( sizeof(option_choice_t) * option_key_num );
  if ( NULL == new_options ) {
    perror("ERROR: ");
    return errno;
  }

  memset( new_options, 0, sizeof(option_choice_t) * option_key_num );

  for ( i = 0; i < option_key_num; i++ ) {
    new_options[i].key = option_key[i];
  }

  status = get_ppd_choice(new_options, ppd);
  if ( 0 != status ) {
    goto er_start_job;
  }

  status = overwrite_option_by_commandline_choice(new_options, ppd, num_options, options);
  if ( 0 != status ) {
    goto er_start_job;
  }

  dbgprintf("\n*** show all options ***\n");
  for ( i = 0; i < option_key_num; i++ ) {
    if ( NULL == new_options[i].choice ) {
      dbgprintf("key=\"%s\"\n", new_options[i].key);
    } else {
      dbgprintf("key=\"%s\", value=\"%s\"\n", new_options[i].key, new_options[i].choice->choice);
    }
  }

  /* init page size parameter */
  for ( i = 0; i > option_key_num; i++ ) {
    if ( 0 == strcmp(PPDKEYPageSize, new_options[i].key) ) {
      break;
    }
  }

  ASSERT( option_key_num != i );
  if ( option_key_num == i ) {
    status = -1;
    goto er_start_job;
  }

  pageSize = ppdPageSize( cast_const_to_ppd_file_t(ppd), new_options[i].choice->choice);
  if ( NULL == pageSize ) {
    log_msg(LOGERROR, "PAGESIZE ERROR.\n");
    status = -1;
    goto er_start_job;
  }

  job_info->options = new_options;
  job_info->page_num = 0;
  job_info->page_size = pageSize;

  
  tmp_cp1 = find_option_choice(PPDKEYPacketMode, option_key_num, job_info->options);
  
  if ( tmp_cp1 ) {
    if ( 0 == strcmp(tmp_cp1, "Auto") ) {
      write_error |= printd_out("\000\000\000\033\001@EJL 1284.4\n@EJL     \n", 27);
      if ( write_error )
      {
          log_msg(LOGERROR, "Can not send cmd. PacketMode...\n");
      }
    }
    dbgprintf("Send PacketMode: %s", tmp_cp1);
  }
  
  return 0;

 er_start_job:

  if ( NULL != new_options ) {
    my_free( new_options );
    new_options = NULL;
  }

  return status;
}

/*
 * End of print...
 */
/*! \brief End Job
  \param  job_info      job information
  \return       error status
  \retval 0     completed
  \retval other error
 */
LOCALFUNC int end_job( job_info_t *job_info )
{
  int write_error = 0;
  unsigned char buf[1];

  ASSERT(NULL != job_info);

  /* reset */
  write_error |= printf_out("%c%s", ESC, "@");

  /* current position move to lefthand. */
  buf[0] = CR;
  write_error |= printd_out(buf, 1);


  fflush(stdout);

  if ( NULL != job_info->input_raster_buff ) {
    my_free(job_info->input_raster_buff);
    job_info->input_raster_buff = NULL;
    job_info->band_buf = NULL;
  }

  cupsRasterClose(job_info->ras);
  job_info->ras = NULL;

  ppdClose(job_info->ppd);
  job_info->ppd = NULL;

  if ( NULL != job_info->options ) {
    my_free( job_info->options );
    job_info->options = NULL;
  }

  if ( write_error ) {
      return EIO;
  }

  return 0;
}


/*! \brief Start Page
  \param  job_info      job information
  \param  ppd           ppd information
  \param  page_header   page information
  \return               error status
  \retval 0             completd
  \retval other         error
 */
LOCALFUNC int start_page( job_info_t *job_info, const ppd_file_t *ppd, const cups_page_header_t *page_header )
{
  unsigned long         page_length   = 0; /* 1/3600 dpi */
  unsigned long         top_margin    = 0; /* 1/3600 dpi */
  unsigned long         bottom_margin = 0; /* 1/3600 dpi */
  unsigned long         left_margin   = 0; /* 1/3600 dpi */
  unsigned long         right_margin  = 0; /* 1/3600 dpi */
  unsigned char         buf[MAX_STR_BUF];
  unsigned long         lnum = 0;	/* temporary value */
  int                   write_error = 0;
  int                   top_shift = 0;
  int                   high_speed = 0;
  int                   left_shift = 0;
  const char           *tmp_cp1 = NULL;
  char                 *tmp_cp2 = NULL;
  double                dvalue = 0;


#if defined(HAVE_SIGACTION) && !defined(HAVE_SIGSET)
  struct sigaction action;                      /* Actions for POSIX signals */
#endif /* HAVE_SIGACTION && !HAVE_SIGSET */

  ASSERT(NULL != job_info);    if ( NULL == job_info )    { return EINVAL; }
  ASSERT(NULL != ppd);         if ( NULL == ppd )         { return EINVAL; }
  ASSERT(NULL != page_header); if ( NULL == page_header ) { return EINVAL; }

  /*
   * Register a signal handler to eject the current page if the
   * job is cancelled.
   */

#ifdef HAVE_SIGSET /* Use System V signals over POSIX to avoid bugs */
  sigset(SIGTERM, cancel_job);
#elif defined(HAVE_SIGACTION)
  memset(&action, 0, sizeof(action));

  sigemptyset(&action.sa_mask);
  action.sa_handler = cancel_job;
  sigaction(SIGTERM, &action, NULL);
#else
  signal(SIGTERM, cancel_job);
#endif /* HAVE_SIGSET */

  if ( 1 != job_info->page_num ) {

    /* current position move to lefthand. */
    buf[0] = CR;
    write_error |= printd_out(buf, 1);

  } else {			/* first page init */

    job_info->input_raster_buff = my_malloc(page_header->cupsBytesPerLine * (page_header->cupsRowCount + 1));
    if ( NULL == job_info->input_raster_buff ) {
      perror("ERROR");
      return ENOMEM;
    }
    job_info->band_buf = &(job_info->input_raster_buff[page_header->cupsBytesPerLine]);


    dbgprintf("\n ******** CUPS page size information (72 dpi) ********\n");
    dbgprintf("width  = %f\n", job_info->page_size->width);
    dbgprintf("length = %f\n", job_info->page_size->length);
    dbgprintf("top    = %f\n", job_info->page_size->top);
    dbgprintf("bottom = %f\n", job_info->page_size->bottom);
    dbgprintf("left   = %f\n", job_info->page_size->left);
    dbgprintf("right  = %f\n", job_info->page_size->right);

    page_length   =  job_info->page_size->length                                * (DESIGNBASE / 72);
    top_margin    = (job_info->page_size->length - job_info->page_size->top   ) * (DESIGNBASE / 72);
    bottom_margin = (job_info->page_size->length - job_info->page_size->bottom) * (DESIGNBASE / 72);
    left_margin   =  job_info->page_size->left                                  * (DESIGNBASE / 72);
    right_margin  =  job_info->page_size->width                                 * (DESIGNBASE / 72);
    if ( (0 > top_margin) || (top_margin > bottom_margin) ) {
      top_margin = 0; /* top margin force set zero */
    }
    dbgprintf("\n ******** page define information (%u dpi) ********\n", DESIGNBASE);
    dbgprintf("page_length   = %lu\n", page_length);
    dbgprintf("top_margin    = %lu\n", top_margin);
    dbgprintf("bottom_margin = %lu\n", bottom_margin);
    dbgprintf("left_margin   = %lu\n", left_margin);
    dbgprintf("right_margin  = %lu\n", right_margin);

    /*
     * get top shift...
     */
    top_shift = 0;
    tmp_cp1 = find_option_choice(PPDKEYVShift, option_key_num, job_info->options);

    if ( tmp_cp1 ) {
      dvalue = strtod(tmp_cp1, &tmp_cp2);

      if ( (TSHIFTMIN > dvalue) ||
	   (TSHIFTMAX < dvalue) ) {
	ASSERT(0);
	return EINVAL;
      }

      top_shift = dvalue * 100 * page_header->HWResolution[1] / (25.4 * 10);

      if ( 0 > top_shift ) {
	if ( -4 > (top_shift % 10) ) {
	  top_shift = top_shift / 10;
	  top_shift = top_shift - 1;
	} else {
	  top_shift = top_shift / 10;
	}
      } else
	if ( 0 < top_shift ) {
	  if ( 4 < (top_shift % 10) ) {
	    top_shift = top_shift / 10;
	    top_shift = top_shift + 1;
	  } else {
	    top_shift = top_shift / 10;
	  }
	}
    }
    job_info->top_shift = top_shift;

    /*
     * get left shift...
     */
    left_shift = 0;
    tmp_cp1 = find_option_choice(PPDKEYHShift, option_key_num, job_info->options);

    if ( tmp_cp1 ) {
      dvalue = strtod(tmp_cp1, &tmp_cp2);

      if ( (LSHIFTMIN > dvalue) ||
	   (LSHIFTMAX < dvalue) ) {
	ASSERT(0);
	return EINVAL;
      }

      left_shift = dvalue * 100 * page_header->HWResolution[0] / (25.4 * 10);

      if ( 0 > left_shift ) {
	if ( -4 > (left_shift % 10) ) {
	  left_shift = left_shift / 10;
	  left_shift = left_shift - 1;
	} else {
	  left_shift = left_shift / 10;
	}
      } else
	if ( 0 < left_shift ) {
	  if ( 4 < (left_shift % 10) ) {
	    left_shift = left_shift / 10;
	    left_shift = left_shift + 1;
	  } else {
	    left_shift = left_shift / 10;
	  }
	}
    }
    job_info->left_shift = left_shift;


    /*
     * Job initialize ESC/P command start...
     */

    /* reset */
    write_error |= printf_out("%c%s", ESC, "@");

    /* current position move to lefthand. */
    buf[0] = CR;
    write_error |= printd_out(buf, 1);

    /* select 10cpi = 1/10 inch */
    write_error |= printf_out("%c%s", ESC, "P");

    tmp_cp1 = find_option_choice(PPDKEYQuality, option_key_num, job_info->options);

    if ( 0 != strcmp(tmp_cp1, "Default") ) {

      /* High-Speed mode */
      if ( NULL != strstr(tmp_cp1, "HS") ) {
	high_speed = 1;		/* High-Speed mode ON */
      } else if ( NULL != strstr(tmp_cp1, "Std") ) {
	high_speed = 0;		/* High-Speed mode OFF */
      } else {
	high_speed = -1;
      }

      if ( -1 != high_speed ) {
	write_error |= printf_out("%c%s", FS, "x");
	buf[0] = high_speed;
	write_error |= printd_out(buf, 1);
       }

      /* turn unidirectional mode on/off */
      write_error |= printf_out("%c%s", ESC, "U");

      if ( NULL != strstr(tmp_cp1, "UniOn") ) {
	buf[0] = 1;		/* unidirectional mode ON */
      } else {
	buf[0] = 0;		/* unidirectional mode OFF */
      }
      write_error |= printd_out(buf, 1);
    }

    /* set unit */
    write_error |= printf_out("%c%s", ESC, "(U");
    buf[0] = 1;
    buf[1] = 0;
    buf[2] = 10;
    write_error |= printd_out(buf, 3); /* base unit = 1/360 inch */

    /* set page length in defined unit */
    write_error |= printf_out("%c%s", ESC, "(C");
    buf[0] = 2;
    buf[1] = 0;

    lnum = my_round_ul(page_length, DESIGNBASE, VMOVEBASE);
    buf[2] = (unsigned char)(lnum % 256);
    buf[3] = (unsigned char)(lnum / 256);
    write_error |= printd_out(buf, 4);

    if ( is_match_key_value(PPDKEYTopMP, "Driver", option_key_num, job_info->options ) ) {

      /* set page format in defined unit */
      write_error |= printf_out("%c%s", ESC, "(c");
      buf[0] = 4;
      buf[1] = 0;

      lnum = my_round_ul(top_margin, DESIGNBASE, VMOVEBASE);
      buf[2] = (unsigned char)(lnum % 256);
      buf[3] = (unsigned char)(lnum / 256);

      lnum = my_round_ul(bottom_margin, DESIGNBASE, VMOVEBASE);
      buf[4] = (unsigned char)(lnum % 256);
      buf[5] = (unsigned char)(lnum / 256);
      write_error |= printd_out(buf, 6);
    }

	/* Print Speed param check. */
	tmp_cp1 = find_option_choice(PPDKEYSpeed, option_key_num, job_info->options);
	if ( NULL != tmp_cp1 ) {
		strncpy(printSpeed, tmp_cp1, strlen(tmp_cp1));
	} else {
		strcpy(printSpeed, "Default");
	}
    dbgprintf("printSpeed:%s\n", printSpeed);

#if 0   /* not support horizontal margin command now. */

    /* set left margin in 1/10 inch */
    write_error |= printf_out("%c%s", ESC, "l");
    lnum = my_round_ul(left_margin, DESIGNBASE, 10);
    buf[0] = (unsigned char)(lnum % 256);
    write_error |= printd_out(buf, 1);


    /* set right margin in 1/10 inch */
    write_error |= printf_out("%c%s", ESC, "Q");
    lnum = my_round_ul(right_margin, DESIGNBASE, 10);
    buf[0] = (unsigned char)(lnum % 256);
    write_error |= printd_out(buf, 1);

#endif	/* not support horizontal margin command now. */

#if 0   /* not support line spacing setting command now. */

    /* set line spacing */
    if ( ( 24 == page_header->cupsRowCount ) ||
	 ( 48 == page_header->cupsRowCount ) ) {

      /* n/360 inch */
      write_error |= printf_out("%c%s", ESC, "+");
      lnum = page_header->cupsRowCount * VMOVEBASE / page_header->HWResolution[1];
      buf[0] = (unsigned char)(lnum % 256);
      write_error |= printd_out(buf, 1);

    } else {

      /* n/180 inch */
      write_error |= printf_out("%c%s", ESC, "3");
      lnum = page_header->cupsRowCount * 180 / page_header->HWResolution[1];
      buf[0] = (unsigned char)(lnum % 256);
      write_error |= printd_out(buf, 1);
    }
#endif  /* not support line spacing setting command now. */

#if 0   /* not support cancel condensed print command now. */

    /* cancel condensed print */
    buf[0] = 0x12;		/* DC2 */
    write_error |= printd_out(buf, 1);

#endif  /* not support cancel condensed print command now. */
  }

  /* view page information */
  dbgprintf("\n ******** page information ********\n");
  dbgprintf("Page=%d\n", job_info->page_num);
  dbgprintf("page width=%d pixel\n", page_header->cupsWidth);
  dbgprintf("page height=%d lines\n", page_header->cupsHeight);
  dbgprintf("band height=%d lines\n", page_header->cupsRowCount);

  /* check output error */
  if ( write_error ) {
    return EIO;
  }

  return 0;
}


/*
 * End of a page printing...
 */

/*! \brief End Page
  \param  job_info      job information
  \return               error status
  \retval 0             completed
  \retval other         error
 */
LOCALFUNC int end_page( job_info_t *job_info )
{
#if defined(HAVE_SIGACTION) && !defined(HAVE_SIGSET)
  struct sigaction action;                      /* Actions for POSIX signals */
#endif /* HAVE_SIGACTION && !HAVE_SIGSET */

  unsigned char             buf[1];
  int                       write_error = 0;

  ASSERT(NULL != job_info);    if ( NULL == job_info )    { return EINVAL; }

  /* form feed */
  buf[0] = FF;
  write_error |= printd_out(buf, 1);

  fflush(stdout);

  if ( NULL != job_info->bitmap_band_buf ) {
    my_free(job_info->bitmap_band_buf);
    job_info->bitmap_band_buf = NULL;
  }

  /*
   * Unregister the signal handler...
   */

#ifdef HAVE_SIGSET /* Use System V signals over POSIX to avoid bugs */
  sigset(SIGTERM, SIG_IGN);
#elif defined(HAVE_SIGACTION)
  memset(&action, 0, sizeof(action));

  sigemptyset(&action.sa_mask);
  action.sa_handler = SIG_IGN;
  sigaction(SIGTERM, &action, NULL);
#else
  signal(SIGTERM, SIG_IGN);
#endif /* HAVE_SIGSET */

  if ( write_error ) {
    return EIO;
  }

  return 0;
}


/*
 * Cancel the current job...
 */

/*! \brief Cancel Job
  \param  sig signal from system (CUPS).
  \return     void
 */
LOCALFUNC void cancel_job(int sig)
{
  long  i;

  dbgprintf("\n ******** JOB CANCEL ******** \n");
  dbgprintf(" signal number=%d\n", sig);

  /*
   * Send out lots of NUL bytes to clear out any pending raster data...
   */
  dbgprintf(" number of send NUL chars=%u\n", job_info->bmp_data_size);
  if ( job_info ) {
    for ( i = 0; i < job_info->bmp_data_size; i++ ) {
      putchar(0);
    }
  }


  /*
   * Send out some NUL bytes for pending ESC command cancel ...
   */
  dbgprintf("\n number of send NUL chars=%d\n", 16);
  for ( i = 0; i < 16; i++ ) {
    putchar(0);
  }

  /*
   * End the current page and exit...
   */

  end_page(job_info);
  end_job(job_info);

  debug_check_memory_allocate_counter();

  exit(0);
}


/*! \brief n line feed
  \param  page_header   page information
  \param  skip_line_num Number of skips
  \return               error status
  \retval 0             completd
  \retval other         error
 */
LOCALFUNC int line_feed( const cups_page_header_t *page_header, unsigned short *skip_line_num )
{
  int                       write_error = 0;
  unsigned short 	    skip_line_180dpi = 0;
  unsigned char             buf[1] = {0,};

  ASSERT(NULL != page_header);   if ( NULL == page_header )   { return EINVAL; }
  ASSERT(NULL != skip_line_num); if ( NULL == skip_line_num ) { return EINVAL; }
  ASSERT(0 != *skip_line_num);   if ( 0 == *skip_line_num )   { return EINVAL; }

  /* paper feed skip_line_num start */
  skip_line_180dpi = my_round_ul(*skip_line_num, page_header->HWResolution[1], 180);
  while ( skip_line_180dpi ) {
    write_error |= printf_out("%c%s", ESC, "J");
    if ( 255 > skip_line_180dpi ) {
      buf[0] = (unsigned char)skip_line_180dpi;
    } else {
      buf[0] = 255;
    }
    write_error |= printd_out(buf, 1);

    skip_line_180dpi = skip_line_180dpi - (unsigned short)buf[0];
  }
  *skip_line_num = 0;
  /* paper feed skip_line_num end */

  if ( write_error ) {
    return EIO;
  }
  return 0;
}

/*! \brief Band data send to printer
  \param  job_info      job information
  \param  page_header   page information
  \param  skip_line_num Number of skips
  \param  band_line     Number of line in the \a band_buf
  \param  band_buf      Band buffer for print image
  \return               error status
  \retval 0             completd
  \retval other         error
 */
LOCALFUNC int send_band( job_info_t *job_info, const cups_page_header_t *page_header, unsigned short *skip_line_num, unsigned short *band_line, unsigned char *band_buf )
{
  int                       ras_width_byte = 0;
  int                       bitmap_band_byte = 0;
  long                      bitmap_buf_size = 0;
  int                       ref_pos = 0;
  int                       set_pos = 0;
  unsigned char             maskbit = 0;
  unsigned char             set_bit = 0;
  int                       x = 0;
  int                       y = 0;
  int                       write_error = 0;
  unsigned char             buf[1] = {0,};
  int                       status = 0;
  int                       black_start_pixel = -1;
  int                       black_end_pixel = -1;

  ASSERT(NULL != job_info);      if ( NULL == job_info )      { return EINVAL; }
  ASSERT(NULL != page_header);   if ( NULL == page_header )   { return EINVAL; }
  ASSERT(NULL != skip_line_num); if ( NULL == skip_line_num ) { return EINVAL; }
  ASSERT(NULL != band_line);     if ( NULL == band_line )     { return EINVAL; }
  ASSERT(NULL != band_buf);      if ( NULL == band_buf )      { return EINVAL; }


  ASSERT(0 != *band_line);       if ( 0 == *band_line )      { return EINVAL; }
  ASSERT(0 == *skip_line_num);

  if ( 32 * 256 <= page_header->cupsWidth ) {
    ASSERT(0);
    return -1;
  }

  ras_width_byte = (page_header->cupsWidth + 7) / 8;
  ASSERT(ras_width_byte == page_header->cupsBytesPerLine);

  bitmap_band_byte = page_header->cupsRowCount / 8;
  ASSERT(0 == page_header->cupsRowCount % 8);

  bitmap_buf_size = ras_width_byte * page_header->cupsRowCount;

  job_info->bitmap_band_buf = my_malloc(bitmap_buf_size);
  if ( NULL == job_info->bitmap_band_buf ) {
    perror("ERROR"); return ENOMEM;
  }

  memset(job_info->bitmap_band_buf, 0, bitmap_band_byte * page_header->cupsWidth);

  black_start_pixel = -1;
  black_end_pixel = -1;

  maskbit = 0;
  set_bit = 0;

  if ( 0 > job_info->left_shift ) {
    x = job_info->left_shift * (-1);
  } else {
    x = 0;
  }

  for ( ; x < page_header->cupsWidth; x++ ) {

    if ( (x + job_info->left_shift) >= page_header->cupsWidth ) {
      break;/* do nothing */
    }

    if ( !maskbit ) {
      maskbit = 0x80;
    }

    for ( y = 0; y < page_header->cupsRowCount; y++ ) {
      ref_pos = y * ras_width_byte + x / 8;
      set_pos = (x + job_info->left_shift) * bitmap_band_byte + y / 8;

      if ( !set_bit ) {
	set_bit = 0x80;
      }

      if ( y < *band_line ) {
	if ( band_buf[ref_pos] & maskbit ) {
	  job_info->bitmap_band_buf[set_pos] |= set_bit;
	  if ( 0 > black_start_pixel ) {
	    black_start_pixel = (x + job_info->left_shift);
	  }
	  black_end_pixel = (x + job_info->left_shift);
	}
      }
      set_bit = set_bit >> 1;
    }
    maskbit = maskbit >> 1;
  }

  ASSERT(0 <= black_end_pixel);
  ASSERT(0 <= black_start_pixel);
  if ( ( 0 <= black_start_pixel) && ( 0 <= black_end_pixel ) ) {
    status = band_out(job_info, job_info->bitmap_band_buf,
		      get_bitimage_mode(page_header->HWResolution[0],
					page_header->HWResolution[1]),
		      bitmap_band_byte,
		      black_start_pixel,
		      black_end_pixel - black_start_pixel + 1,
		      page_header->HWResolution[0]);
  }

  my_free(job_info->bitmap_band_buf);
  job_info->bitmap_band_buf = NULL;

  if ( 0 != status ) {
    return status;
  }
  /* current position move to lefthand. */
  buf[0] = CR;
  write_error |= printd_out(buf, 1);
  *skip_line_num = *band_line;
  *band_line = 0;
  /* band data send to printer end */

  if ( write_error ) {
    return EIO;
  }
  return 0;
}


/*! \brief Output one band bitmap image data to printer.
  \param  job_info      job information
  \param  bmpbuf        bitmap image data buffer
  \param  mode          bitimage mode (resolution)
  \param  band_bytes    bitmap image data height (in Bytes)
  \param  start         black bitmap image start x point (in pixel)
  \param  length        black bitmap image length (in pixel)
  \param  xreso         horizontal resolution the bitmap image data
  \return               error status
  \retval 0             completed
  \retval other         error
 */
LOCALFUNC int band_out( job_info_t *job_info, unsigned char *bmpbuf, int mode, int band_bytes, int start, int length, int xreso )
{
  unsigned char     buf[4] = {0,};
  int               write_error = 0;
  unsigned char    *sendbuf = NULL;
  long              lnum = 0;

  ASSERT(NULL != job_info);    if ( NULL == job_info )   { return EINVAL; }
  ASSERT(NULL != bmpbuf);      if ( NULL == bmpbuf )     { return EINVAL; }
  ASSERT(0 <= mode);           if ( 0 > mode )           { return EINVAL; }
  ASSERT(0 < band_bytes);      if ( 0 >= band_bytes )    { return EINVAL; }
  ASSERT(0 <= start);          if ( 0 > start )          { return EINVAL; }
  ASSERT(0 < length);          if ( 0 >= length )        { return EINVAL; }
  ASSERT(32 * 256 > length);   if ( 32 * 256 <= length ) { return EINVAL; }
  ASSERT(0 < xreso);           if ( 0 >= xreso )         { return EINVAL; }
  sendbuf = &bmpbuf[band_bytes * start];
  lnum = my_round_ul(start, xreso, HMOVEBASE);


  if ( 0 != start ) {
    write_error |= printf_out("%c%s", ESC, "$");
    buf[0] = lnum % 256;
    buf[1] = lnum / 256;
    write_error |= printd_out(buf, 2);
  }


  job_info->bmp_data_size = band_bytes * length; /* for job cancel */

  write_error |= printf_out("%c%s", ESC, "*");
  buf[0] = mode;
  buf[1] = length % 256;
  buf[2] = length / 256;
  write_error |= printd_out(buf, 3);

  write_error |= printd_out(sendbuf, band_bytes * length);

  job_info->bmp_data_size = 0; /* for job cancel */

  return write_error;
}

/*! \brief Get bit image mode
  \param  xreso  horizontal resolution
  \param  yreso  vertical resolution
  \return        bitimage mode number.
  \retval -1     error
 */
LOCALFUNC int get_bitimage_mode( unsigned int xreso, unsigned int yreso )
{
  int mode = -1;

  if ( 60 == yreso ) {
    if ( 60 == xreso ) { mode = 0; } else
      if ( 120 == xreso ) { mode = 1; } else
	if ( 240 == xreso ) { mode = 3; } else
	  if ( 80 == xreso ) { mode = 4; } else
	    if ( 90 == xreso ) { mode = 6; }

  } else
    if( 180 == yreso ) {
      if ( 60 == xreso ) { mode = 32; } else
	if ( 120 == xreso ) { mode = 33; } else
	  if ( 90 == xreso ) { mode = 38; } else
	    if ( 180 == xreso ) { mode = 39; } else
	      if ( 360 == xreso ) { mode = 40; }
    } else
      if( 360 == yreso ) {
	if ( 180 == xreso ) { mode = 71; } else
	  if ( 360 == xreso ) { mode = 72; }
      }
	
	// check printSpeed val.
	if (  0 == strcmp( printSpeed, "EpspdHigh" )) {
		//High speed x (120x180) => mode:34
		if ( ( 120 == xreso ) && (180 == yreso) ) { mode = 34;} else
		//High speed x (180x180) => mode:37
		if ( ( 180 == xreso ) && (180 == yreso) ) { mode = 37;} 
	}
    dbgprintf("Reso mode  = %d[0x%02x]\n", mode, mode);
	
  return mode;
}


/*
 * Sub functions...
 */

/*
 * for adjust CUPS API
 */

/*! \brief Cast function
  \param  p   const pointer
  \return     non const pointer
 */
LOCALFUNC ppd_file_t *cast_const_to_ppd_file_t ( const ppd_file_t *p )
{
  return (ppd_file_t*)p;
}

/*! \brief Cast function
  \param  p   const pointer
  \return     non const pointer
 */
LOCALFUNC ppd_option_t *cast_const_to_ppd_option_t( const ppd_option_t *p )
{
  return (ppd_option_t*)p;
}




/*! \brief Find value in the option list
  \param key
  \param num_options
  \param options
  \return            found value for \a key.
  \retval NULL       not found the \a key in \a options.
 */
LOCALFUNC const char *find_option_value( const char *key, const int num_options, const cups_option_t *options )
{
  int i = 0;

  ASSERT(NULL != key);     if ( NULL == key )     { return NULL; }
  ASSERT(NULL != options); if ( NULL == options ) { return NULL; }
  ASSERT(num_options);     if ( !num_options )    { return NULL; }

  for ( i = 0; i < num_options; i++ ) {
    if ( 0 == strcmp(key, options[i].name) ) {
      return options[i].value;
    }
  }
  return NULL;
}

/*! \brief Find option choice
  \param key
  \param num_options
  \param options
  \return            found value for \a key.
  \retval NULL       not found the \a key in \a options.
 */
LOCALFUNC const char *find_option_choice( const char *key, int num_options, const option_choice_t *options )
{
  int i = 0;

  ASSERT(NULL != key);     if ( NULL == key )     { return NULL; }
  ASSERT(NULL != options); if ( NULL == options ) { return NULL; }
  ASSERT(num_options);     if ( !num_options )    { return NULL; }

  for ( i = 0; i < num_options; i++ ) {
    if ( 0 == strcmp(key, options[i].key) ) {
      if ( options[i].choice ) {
        return options[i].choice->choice;
      } else {
        return NULL;
      }
    }
  }
  return NULL;
}

/*! \brief Check value in the option list.
  \param key
  \param value
  \param num_options
  \param options
  \return            found status.
  \retval 1       found
  \retval 0       not found
 */
LOCALFUNC int is_match_key_value( const char *key, const char *value, int num_options, const option_choice_t *options )
{
  const char *found_value = NULL;

  ASSERT(NULL != key);     if ( NULL == key )     { return 0; }
  ASSERT(NULL != value);   if ( NULL == value )   { return 0; }
  ASSERT(NULL != options); if ( NULL == options ) { return 0; }
  ASSERT(num_options);     if ( !num_options )    { return 0; }

  found_value = find_option_choice(key, num_options, options);
  if ( NULL != found_value ) {
    if ( 0 == strcmp(value, found_value) ) {
      return 1;			/* matched */
    }
  }
  return 0;			/* un-matched */
}


/*! \brief Get ppd choice
  \param new_options
  \param ppd
  \return     error status
  \retval  0  no error
  \retval -1  error
 */
LOCALFUNC int get_ppd_choice( option_choice_t *new_options, const ppd_file_t *ppd )
{
  const ppd_option_t *option = NULL;
  const ppd_choice_t *choice = NULL;
  int i = 0;

  ASSERT(NULL != new_options); if ( NULL == new_options ) { return EINVAL; }
  ASSERT(NULL != ppd);         if ( NULL == ppd )         { return EINVAL; }

  for ( i = 0; i < option_key_num; i++ ) {

    dbgprintf("No.%d, key=%s\n", i, new_options[i].key);

    choice = ppdFindMarkedChoice( cast_const_to_ppd_file_t(ppd), new_options[i].key );

    if ( NULL == choice ) {

      /* Can't find marked option in the PPD file. */
      option = ppdFindOption( cast_const_to_ppd_file_t(ppd), new_options[i].key );

      if ( NULL == option ) {
        if ( ( 0 == strcmp(PPDKEYSpeed, new_options[i].key) ) ||
             ( 0 == strcmp(PPDKEYPacketMode, new_options[i].key) ) ) {
          /* not necessary ppd key. */
          dbgprintf("Can't find ppdFindOption(). not necessary key:%s.\n", new_options[i].key);
          continue;
        } else {
          log_msg(LOGERROR, "PPD file option key (%s) error.\n", new_options[i].key );
          return -1;
        }
      } else {

        choice = ppdFindChoice( cast_const_to_ppd_option_t(option), option->defchoice );

        if ( NULL == choice ) {
          log_msg(LOGERROR, "PPD file option value (%s) error.\n", option->defchoice);
          return -1;
        }
      }
    }

   new_options[i].choice = choice;
   dbgprintf("find key=%s\n", new_options[i].key);
  }

  return 0;
}

/*! \brief Overwrite current option setting by commandline parameter
  \param new_options	new option list.
  \param ppd		ppd file parameter in CUPS.
  \param num_options	number of items in \a options.
  \param options	CUPS option list.
  \return     error status
  \retval  0  no error
  \retval -1  error
 */
LOCALFUNC int overwrite_option_by_commandline_choice( option_choice_t *new_options, const ppd_file_t *ppd, const int num_options, const cups_option_t *options )
{
  const ppd_option_t *option = NULL;
  const ppd_choice_t *choice = NULL;
  const char         *value = NULL;
  int i;

  ASSERT(NULL != new_options); if ( NULL == new_options ) { return EINVAL; }
  ASSERT(NULL != ppd);         if ( NULL == ppd )         { return EINVAL; }

  if ( 0 == num_options ) {
    return 0;			/* no commandline option */
  }

  ASSERT(NULL != options);     if ( NULL == options )     { return EINVAL; }

  for ( i = 0; i < option_key_num; i++ ) {

    value = find_option_value( new_options[i].key, num_options, options );

    if ( NULL != value ) {

      option = ppdFindOption( cast_const_to_ppd_file_t(ppd), new_options[i].key );

      if ( NULL == option ) {
        log_msg(LOGERROR, "option key (%s) error.\n", new_options[i].key);
        return -1;
      }

      choice = ppdFindChoice( cast_const_to_ppd_option_t(option), value );

      if ( NULL == choice ) {
        log_msg(LOGERROR, "option key (%s) value (%s) error.\n", new_options[i].key, value);
        return -1;
      }

      new_options[i].choice = choice;

    }
  }
  return 0;
}

/* end of file */
